# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.database.entrez"
__author__ = "Patrick Kunzmann"
__all__ = ["Query", "SimpleQuery", "CompositeQuery", "search"]

import requests
import abc
from xml.etree import ElementTree
from .check import check_for_errors


_base_url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/"

_search_url = ("esearch.fcgi?db={:}"
              "&term={:}"
              "&retmax={:}")

class Query(metaclass=abc.ABCMeta):
    """
    Base class for a wrapper around a search term
    for the NCBI Entrez search service.
    """
    def __init__(self):
        pass
    
    @abc.abstractmethod
    def __str__(self):
        pass
    
    def __or__(self, operand):
        if not isinstance(operand, Query):
            operand = SimpleQuery(operand)
        return CompositeQuery("OR", self, operand)
    
    def __and__(self, operand):
        if not isinstance(operand, Query):
            operand = SimpleQuery(operand)
        return CompositeQuery("AND", self, operand)
    
    def __xor__(self, operand):
        if not isinstance(operand, Query):
            operand = SimpleQuery(operand)
        return CompositeQuery("NOT", self, operand)


class CompositeQuery(Query):
    """
    A representation of an composite query
    for the NCBI Entrez search service.
    
    A composite query is a combination of two other queries,
    combined either with an 'AND', 'OR' or 'NOT' operator.

    Usually the user does not create instances of this class directly,
    but :class:`Query` instances are combined with
    ``|`` (OR), ``&`` (AND) or ``^`` (NOT).
    
    Parameters
    ----------
    operator: str, {"AND", "OR", "NOT"}
        The combination operator.
    queries : iterable object of SimpleQuery
        The queries to be combined.
    
    Examples
    --------

    >>> query = SimpleQuery("Escherichia coli", "Organism") & \\
    ... SimpleQuery("90:100", "Sequence Length")
    >>> print(type(query).__name__)
    CompositeQuery
    >>> print(query)
    ("Escherichia coli"[Organism]) AND (90:100[Sequence Length])
    """
    
    def __init__(self, operator, query1, query2):
        super().__init__()
        self._op = operator
        self._q1 = query1
        self._q2 = query2
    
    def __str__(self):
        return "({:}) {:} ({:})".format(str(self._q1), self._op, self._q2)
        


class SimpleQuery(Query):
    """
    A simple query for the NCBI Entrez search service without
    combination via 'AND', 'OR' or 'NOT'. A query consists of a search
    term and an optional field.

    A list of available search fields with description can be found
    `here <https://www.ncbi.nlm.nih.gov/books/NBK49540/>`_.
    
    Parameters
    ----------
    term: str
        The search term.
    field : str, optional
        The field to search the term in.
        The list of possible fields and the required search term
        formatting can be found
        `here <https://www.ncbi.nlm.nih.gov/books/NBK49540/>`_.
        By default the field is omitted and all fields are searched in
        for the term, implicitly.
    
    Examples
    --------
    
    >>> query = SimpleQuery("Escherichia coli")
    >>> print(query)
    "Escherichia coli"
    >>> query = SimpleQuery("Escherichia coli", "Organism")
    >>> print(query)
    "Escherichia coli"[Organism]
    """

    # Field identifiers are taken from
    # https://www.ncbi.nlm.nih.gov/books/NBK49540/
    _fields = [
        "Accession", "All Fields", "Author", "EC/RN Number", "Feature Key",
        "Filter", "Gene Name", "Genome Project", "Issue", "Journal", "Keyword",
        "Modification Date", "Molecular Weight", "Organism", "Page Number",
        "Primary Accession", "Properties", "Protein Name", "Publication Date",
        "SeqID String", "Sequence Length", "Substance Name", "Text Word",
        "Title", "Volume",
        # Abbreviations
        "ACCN", "ALL", "AU", "AUTH", "ECNO", "FKEY", "FILT", "SB", "GENE",
        "ISS", "JOUR", "KYWD", "MDAT", "MOLWT", "ORGN", "PAGE", "PACC",
        "PORGN", "PROP", "PROT", "PDAT", "SQID", "SLEN", "SUBS", "WORD",  "TI",
        "TITL" "VOL"
    ]

    def __init__(self, term, field=None):
        super().__init__()
        if field is not None:
            if field not in SimpleQuery._fields:
                raise ValueError(f"Unknown field identifier '{field}'")
        for invalid_string in \
            ['"', "AND", "OR", "NOT", "[", "]", "(", ")", "\t", "\n"]:
                if invalid_string in term:
                    raise ValueError(
                        f"Query contains illegal term {invalid_string}"
                    )
        if " " in term:
            # Encapsulate in quotes if spaces are in search term
            term = f'"{term}"'
        self._term = term
        self._field = field
    
    def __str__(self):
        string = self._term
        if self._field is not None:
            string += f"[{self._field}]"
        return string


def search(query, db_name, number=20):
    r"""
    Get all PDB IDs that meet the given query requirements,
    via the NCBI ESearch service.
    
    This function requires an internet connection.
    
    Parameters
    ----------
    query : Query
        The search query.
    db_name : str:
        E-utility database name.
    number : Query
        The maximum number of UIDs that are obtained.
    
    Returns
    -------
    ids : list of str
        A list of strings containing all NCBI UIDs (accession number)
        that meet the query requirements.
    
    Warnings
    --------
    Even if you give valid input to this function, in rare cases the
    database might return no or malformed data to you.
    In these cases the request should be retried.
    When the issue occurs repeatedly, the error is probably in your
    input.

    Notes
    -----
    A list of available search fields with description can be found
    `here <https://www.ncbi.nlm.nih.gov/books/NBK49540/>`_.
    
    Examples
    --------
    >>> query = SimpleQuery("Escherichia coli", "Organism") & \
    ...         SimpleQuery("90:100", "Sequence Length")
    >>> ids = search(query, "nuccore", number=5)
    >>> print(ids)
    ['...', '...', '...', '...', '...']
    """ 
    r = requests.get(
        (_base_url + _search_url).format(db_name, str(query), str(number))
    )
    xml_response = r.text
    check_for_errors(xml_response)
    root = ElementTree.fromstring(xml_response)
    xpath = ".//IdList/Id"
    uids = [element.text for element in root.findall(xpath)]
    return uids
    


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import biotite.structure as struc
import biotite.structure.io as strucio
import numpy as np
from os.path import join
from ..util import data_dir
import pytest


@pytest.fixture
def array():
    return strucio.load_structure(join(data_dir("structure"), "1l2y.mmtf"))[0]


def test_apply_residue_wise(array):
    data = struc.apply_residue_wise(array, np.ones(len(array)), np.sum)
    assert data.tolist() == [len(array[array.res_id == i])
                             for i in range(1, 21)]

def test_spread_residue_wise(array):
    input_data = np.arange(1,21)
    output_data = struc.spread_residue_wise(array, input_data)
    assert output_data.tolist() == array.res_id.tolist()


def test_get_residue_masks(array):
    SAMPLE_SIZE = 100
    np.random.seed(0)
    indices = np.random.randint(0, array.array_length(), SAMPLE_SIZE)
    masks = struc.get_residue_masks(array, indices)
    for index, mask in zip(indices, masks):
        ref_mask = array.res_id == array.res_id[index]
        assert mask.tolist() == ref_mask.tolist()


def test_get_residue_starts_for(array):
    SAMPLE_SIZE = 100
    np.random.seed(0)
    indices = np.random.randint(0, array.array_length(), SAMPLE_SIZE)
    ref_starts = np.array(
        [np.where(mask)[0][0] for mask
         in struc.get_residue_masks(array, indices)]
    )
    test_starts = struc.get_residue_starts_for(array, indices)
    assert test_starts.tolist() == ref_starts.tolist()


def test_get_residues(array):
    ids, names = struc.get_residues(array)
    assert ids.tolist() == list(range(1, 21))
    assert names.tolist() == ["ASN","LEU","TYR","ILE","GLN","TRP","LEU","LYS",
                              "ASP","GLY","GLY","PRO","SER","SER","GLY","ARG",
                              "PRO","PRO","PRO","SER"]
    assert len(ids) == struc.get_residue_count(array)


def test_residue_iter(array):
    centroid = [struc.centroid(res).tolist()
                for res in struc.residue_iter(array)]
    ref_centroid = struc.apply_residue_wise(
        array, array.coord, np.average, axis=0
    )
    assert centroid == ref_centroid.tolist()


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.database.entrez"
__author__ = "Patrick Kunzmann"
__all__ = ["get_database_name", "fetch", "fetch_single_file"]

from os.path import isdir, isfile, join, getsize
import os
import glob
import io
import requests
from .check import check_for_errors
from ..error import RequestError


_fetch_url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi"


_databases = {"BioProject"        : "bioproject",
              "BioSample"         : "biosample",
              "Biosystems"        : "biosystems",
              "Books"             : "books",
              "Conserved Domains" : "cdd",
              "dbGaP"             : "gap",
              "dbVar"             : "dbvar",
              "Epigenomics"       : "epigenomics",
              "EST"               : "nucest",
              "Gene"              : "gene",
              "Genome"            : "genome",
              "GEO Datasets"      : "gds",
              "GEO Profiles"      : "geoprofiles",
              "GSS"               : "nucgss",
              "HomoloGene"        : "homologene",
              "MeSH"              : "mesh",
              "NCBI C++ Toolkit"  : "toolkit",
              "NCBI Web Site"     : "ncbisearch",
              "NLM Catalog"       : "nlmcatalog",
              "Nucleotide"        : "nuccore",
              "OMIA"              : "omia",
              "PopSet"            : "popset",
              "Probe"             : "probe",
              "Protein"           : "protein",
              "Protein Clusters"  : "proteinclusters",
              "PubChem BioAssay"  : "pcassay",
              "PubChem Compound"  : "pccompound",
              "PubChem Substance" : "pcsubstance",
              "PubMed"            : "pubmed",
              "PubMed Central"    : "pmc",
              "SNP"               : "snp",
              "SRA"               : "sra",
              "Structure"         : "structure",
              "Taxonomy"          : "taxonomy",
              "UniGene"           : "unigene",
              "UniSTS"            : "unists"}

def get_database_name(database):
    """
    Map a common NCBI Entrez database name to an E-utility database
    name.
    
    Parameters
    ----------
    database : str
        Entrez database name.
    
    Returns
    -------
    name : str
        E-utility database name.
    
    Examples
    --------
    
    >>> print(get_database_name("Nucleotide"))
    nuccore
    """
    return _databases[database]


def fetch(uids, target_path, suffix, db_name, ret_type,
          ret_mode="text", overwrite=False, verbose=False):
    """
    Download files from the NCBI Entrez database in various formats.
    
    The data for each UID will be fetched into a separate file.
    
    A list of valid database, retrieval type and mode combinations can
    be found under
    `<https://www.ncbi.nlm.nih.gov/books/NBK25499/table/chapter4.T._valid_values_of__retmode_and/?report=objectonly>`_
    
    This function requires an internet connection.
    
    Parameters
    ----------
    uids : str or iterable object of str
        A single *unique identifier* (UID) or a list of UIDs of the
        file(s) to be downloaded.
    target_path : str or None
        The target directory of the downloaded files.
        If ``None``, the file content is stored in a file-like object
        (`StringIO` or `BytesIO`, respectively).
    suffix : str
        The file suffix of the downloaded files. This value is
        independent of the retrieval type.
    db_name : str:
        E-utility or common database name.
    ret_type : str
        Retrieval type
    ret_mode : str, optional
        Retrieval mode
    overwrite : bool, optional
        If true, existing files will be overwritten. Otherwise the
        respective file will only be downloaded if the file does not
        exist yet in the specified target directory or if the file is
        empty. (Default: False)
    verbose: bool, optional
        If true, the function will output the download progress.
        (Default: False)
    
    Returns
    -------
    files : str or StringIO or BytesIO or list of (str or StringIO or BytesIO)
        The file path(s) to the downloaded files.
        If a single string (a single UID) was given in `uids`,
        a single string is returned. If a list (or other iterable
        object) was given, a list of strings is returned.
        If `target_path` is ``None``, the file contents are stored in
        either `StringIO` or `BytesIO` objects.
    
    Warnings
    --------
    Even if you give valid input to this function, in rare cases the
    database might return no or malformed data to you.
    In these cases the request should be retried.
    When the issue occurs repeatedly, the error is probably in your
    input.
    
    See also
    --------
    fetch_single_file
    
    Examples
    --------
    
    >>> import os.path
    >>> files = fetch(["1L2Y_A","3O5R_A"], path_to_directory, suffix="fa",
    ...               db_name="protein", ret_type="fasta")
    >>> print([os.path.basename(file) for file in files])
    ['1L2Y_A.fa', '3O5R_A.fa']
    """
    # If only a single UID is present,
    # put it into a single element list
    if isinstance(uids, str):
        uids = [uids]
        single_element = True
    else:
        single_element = False
    # Create the target folder, if not existing
    if target_path is not None and not isdir(target_path):
        os.makedirs(target_path)
    files = []
    for i, id in enumerate(uids):
        # Verbose output
        if verbose:
            print(f"Fetching file {i+1:d} / {len(uids):d} ({id})...", end="\r")
        # Fetch file from database
        if target_path is not None:
            file = join(target_path, id + "." + suffix)
        else:
            file = None
        if file is None \
           or not isfile(file) \
           or getsize(file) == 0 \
           or overwrite:
                param_dict = {
                    "db" : _sanitize_db_name(db_name),
                    "id" : id,
                    "rettype" : ret_type,
                    "retmode" : ret_mode,
                    "tool" : "Biotite",
                    "mail" : "padix.key@gmail.com"
                }
                r = requests.get(_fetch_url, params=param_dict)
                content = r.text
                check_for_errors(content)
                if content.startswith(" Error"):
                    raise RequestError(content[8:])
                if file is None:
                    file = io.StringIO(content)
                else:
                    with open(file, "w+") as f:
                        f.write(content)
        files.append(file)
    if verbose:
        print("\nDone")
    # If input was a single ID, return only a single path
    if single_element:
        return files[0]
    else:
        return files


def fetch_single_file(uids, file_name, db_name, ret_type, ret_mode="text",
                      overwrite=False):
    """
    Almost the same as :func:`fetch()`, but the data for the given UIDs
    will be stored in a single file.
    
    Parameters
    ----------
    uids : iterable object of str
        A list of UIDs of the
        file(s) to be downloaded.
    file_name : str or None
        The file path, including file name, to the target file.
    db_name : str:
        E-utility or common database name.
    ret_type : str
        Retrieval type.
    ret_mode : str, optional
        Retrieval mode.
    overwrite : bool, optional
        If false, the file is only downloaded, if no file with the same
        name already exists.
    
    Returns
    -------
    file : str or StringIO or BytesIO
        The file name of the downloaded file.
        If `file_name` is ``None``, the file content is stored in
        either a `StringIO` or a `BytesIO` object.
    
    Warnings
    --------
    Even if you give valid input to this function, in rare cases the
    database might return no or malformed data to you.
    In these cases the request should be retried.
    When the issue occurs repeatedly, the error is probably in your
    input.
    
    See also
    --------
    fetch
    """
    if file_name is not None \
       and os.path.isfile(file_name) \
       and getsize(file_name) > 0 \
       and not overwrite:
            # Do no redownload the already existing file
            return file_name
    uid_list_str = ""
    for id in uids:
        uid_list_str += id + ","
    # Remove terminal comma
    uid_list_str = uid_list_str[:-1]
    param_dict = {
        "db" : _sanitize_db_name(db_name),
        "id" : uid_list_str,
        "rettype" : ret_type,
        "retmode" : ret_mode,
        "tool" : "Biotite",
        "mail" : "padix.key@gmail.com"
    }
    r = requests.get(_fetch_url, params=param_dict)
    content = r.text
    check_for_errors(content)
    if content.startswith(" Error"):
        raise RequestError(content[8:])
    if file_name is None:
        return io.StringIO(content)
    else:
        with open(file_name, "w+") as f:
            f.write(content)
        return file_name


def _sanitize_db_name(db_name):
    if db_name in _databases.keys():
        # Convert into E-utility database name
        return _databases[db_name]
    elif db_name in _databases.values():
        # Is already E-utility database name
        return db_name
    else:
        raise ValueError("Database '{db_name}' is not existing")


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence.graphics"
__author__ = "Patrick Kunzmann"
__all__ = ["plot_dendrogram"]

import numpy as np

def plot_dendrogram(axes, tree, orientation="left", use_distances=True,
                    labels=None, label_size=None, color="black",
                    show_distance=True, **kwargs):
    """
    Plot a dendrogram from a (phylogenetic) tree.

    Parameters
    ----------
    tree : Tree
        The tree to be visualized
    orientation : {'left', 'right', 'bottom', 'top'}, optional
        The position of the root node in the plot
    use_distances : bool, optional
        If true, the `distance` attribute of the :class:`TreeNode`
        objects are used as distance measure.
        Otherwise the topological distance is used.
    labels : list of str, optional 
        The leaf node labels.
        The label of a leaf node is the entry at the position of its
        `index` attribute.
    label_size : float, optional
        The font size of the labels
    color : tuple or str, optional
        A *Matplotlib* compatible color, that is used to draw the lines
        of the dendrogram.
    show_distance : bool, optional
        If true, the distance from the root is shown on the
        corresponding axis.
    **kwargs
        Additional parameters that are used to draw the dendrogram
        lines.
    """
    
    indices = tree.root.get_indices()
    leaf_dict = {indices[i] : i for i in indices}

    # Required for setting the plot limits
    max_distance = 0

    def _plot_node(node, distance):
        """
        Draw the lines from the given node to its children.
        
        Parameters
        ----------
        dist : float
            the distance of the node from root
        
        Returns
        -------
        pos : float
            the postion of the node on the 'label' axis
        """
        # The term 'distance'
        # refers to positions along the 'distance' axis
        # the term 'pos'
        # refers to positions along the other axis
        nonlocal max_distance
        if max_distance < distance:
            max_distance = distance
        if node.is_leaf():
            # No children -> no line can be drawn
            return leaf_dict[node.index]
        else:
            children = node.children
            if use_distances:
                child_distances = [distance + c.distance for c in children]
            else:
                # Use topologic distance of children to this node,
                # which is always 1
                child_distances = [distance + 1 for c in children]
            child_pos = [
                _plot_node(child, child_distance)
                for child, child_distance in zip(children, child_distances)
            ]
            # Position of this node is in the center of the child nodes
            center_pos = sum(child_pos) / len(child_pos)
            if orientation in ["left", "right"]:
                # Line connecting the childs
                axes.plot(
                    [distance, distance], [child_pos[0], child_pos[-1]],
                    color=color, marker="None", **kwargs
                )
                # Lines depicting the distances of the childs
                for child_dist, pos in zip(child_distances, child_pos):
                    axes.plot(
                        [distance, child_dist], [pos, pos],
                        color=color, marker="None", **kwargs
                    )
            elif orientation in ["bottom", "top"]:
                # Line connecting the childs
                axes.plot(
                    [child_pos[0], child_pos[-1]], [distance, distance],
                    color=color, marker="None", **kwargs
                )
                # Lines depicting the distances of the childs
                for child_dist, pos in zip(child_distances, child_pos):
                    axes.plot(
                        [pos, pos], [distance, child_dist],
                        color=color, marker="None", **kwargs
                    )
            else:
                raise ValueError(f"'{orientation}' is not a valid orientation")
            return center_pos
    
    _plot_node(tree.root, 0)

    if labels is not None:
        # Sort labels using the order of indices in the tree
        # A list cannot be directly indexed with a list,
        # hence the conversion to a ndarray
        labels = np.array(labels)[indices].tolist()
    else:
        labels = [str(i) for i in indices]
    # The distance axis does not start at 0,
    # since the root line would not properly rendered
    # Hence the limit is set a to small fraction of the entire axis
    # beyond 0
    zero_limit = -0.01 * max_distance
    if orientation == "left":
        axes.set_xlim(zero_limit, max_distance)
        axes.set_ylim(-1, len(indices))
        axes.set_yticks(np.arange(0, len(indices)))
        axes.set_yticklabels(labels)
        axes.yaxis.set_tick_params(
            left=False, right=False, labelleft=False, labelright=True,
            labelsize=label_size
        )
        axes.xaxis.set_tick_params(
            bottom=True, top=False, labelbottom=show_distance, labeltop=False,
            labelsize=label_size
        )
    elif orientation == "right":
        axes.set_xlim(max_distance, zero_limit)
        axes.set_ylim(-1, len(indices))
        axes.set_yticks(np.arange(0, len(indices)))
        axes.set_yticklabels(labels)
        axes.yaxis.set_tick_params(
            left=False, right=False, labelleft=True, labelright=False,
            labelsize=label_size
        )
        axes.xaxis.set_tick_params(
            bottom=True, top=False, labelbottom=show_distance, labeltop=False,
            labelsize=label_size
        )
    elif orientation == "bottom":
        axes.set_ylim(zero_limit, max_distance)
        axes.set_xlim(-1, len(indices))
        axes.set_xticks(np.arange(0, len(indices)))
        axes.set_xticklabels(labels)
        axes.xaxis.set_tick_params(
            bottom=False, top=False, labelbottom=False, labeltop=True,
            labelsize=label_size
        )
        axes.yaxis.set_tick_params(
            left=True, right=False, labelleft=show_distance, labelright=False,
            labelsize=label_size
        )
    elif orientation == "top":
        axes.set_ylim(max_distance, zero_limit)
        axes.set_xlim(-1, len(indices))
        axes.set_xticks(np.arange(0, len(indices)))
        axes.set_xticklabels(labels)
        axes.xaxis.set_tick_params(
            bottom=False, top=False, labelbottom=True, labeltop=False,
            labelsize=label_size
        )
        axes.yaxis.set_tick_params(
            left=True, right=False, labelleft=show_distance, labelright=False,
            labelsize=label_size
        )
    else:
        raise ValueError(f"'{orientation}' is not a valid orientation")
    axes.set_frame_on(False)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from os.path import join, dirname, realpath
import urllib.error
import urllib.request
import importlib
import shutil


def data_dir(subdir):
    return join(dirname(realpath(__file__)), subdir, "data")


### Functions for conditional test skips ###

tested_urls = {}
def cannot_connect_to(url):
    if url not in tested_urls:
        try:
            urllib.request.urlopen(url)
            tested_urls[url] = False
        except urllib.error.URLError:
            tested_urls[url] = True
    return tested_urls[url]

def cannot_import(module):
    return importlib.util.find_spec(module) is None

def is_not_installed(program):
    return shutil.which(program) is None


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence"
__author__ = "Patrick Kunzmann"
__all__ = ["Location", "Feature", "Annotation", "AnnotatedSequence"]

import numbers
import copy
import sys
from enum import Flag, Enum, auto
import numpy as np
from .sequence import Sequence
from ..copyable import Copyable
from .seqtypes import NucleotideSequence


class Location():
    """
    A :class:`Location` defines at which base(s)/residue(s) a feature is
    located.

    A feature can have multiple :class:`Location` instances if multiple
    locations are joined.

    Objects of this class are immutable.

    Attributes
    ----------
    first : int
        Starting base or residue position of the feature.
    last : int
        Inclusive ending base or residue position of the feature.
    strand : Strand
        The strand direction.
        Always :attr:`Strand.FORWARD` for peptide features.
    defect : Defect
        A possible defect of the location.
    """

    class Defect(Flag):
        """
        This enum type describes location defects.

        A location has a defect, when the feature itself is not directly
        located in the range of the first to the last base.

           - **NONE** - No location defect
           - **MISS_LEFT** - A part of the feature has been truncated
             before the first base/residue of the :class:`Location`
             (probably by indexing an :class:`Annotation` object)
           - **MISS_RIGHT** - A part of the feature has been truncated
             after the last base/residue of the :class:`Location`
             (probably by indexing an :class:`Annotation` object)
           - **BEYOND_LEFT** - The feature starts at an unknown position
             before the first base/residue of the :class:`Location`
           - **BEYOND_RIGHT** - The feature ends at an unknown position
             after the last base/residue of the :class:`Location`
           - **UNK_LOC** - The exact position is unknown, but it is at a
             single base/residue between the first and last residue of
             the :class:`Location`, inclusive
           - **BETWEEN** - The position is between to consecutive
             bases/residues.
        """
        NONE         = 0
        MISS_LEFT    = auto()
        MISS_RIGHT   = auto()
        BEYOND_LEFT  = auto()
        BEYOND_RIGHT = auto()
        UNK_LOC      = auto()
        BETWEEN      = auto()

    class Strand(Enum):
        """
        This enum type describes the strand of the feature location.
        This is not relevant for protein sequence features.
        """
        FORWARD = auto()
        REVERSE = auto()

    def __init__(self, first, last, strand=Strand.FORWARD,
                 defect=Defect.NONE):
        if first > last:
            raise ValueError(
                "The first position cannot be higher than the last position"
            )
        self._first = first
        self._last = last
        self._strand = strand
        self._defect = defect

    def __repr__(self):
        """Represent Location as a string for debugging."""
        return f'Location({self._first}, {self._last}, strand={"Location." + str(self._strand)}, ' \
               f'defect={"Location." + str(self._defect)})'

    @property
    def first(self):
        return self._first

    @property
    def last(self):
        return self._last

    @property
    def strand(self):
        return self._strand

    @property
    def defect(self):
        return self._defect

    def __str__(self):
        string = "{:d}-{:d}".format(self.first, self.last)
        if self.strand == Location.Strand.FORWARD:
            string = string + " >"
        else:
            string = "< " + string
        return string

    def __eq__(self, item):
        if not isinstance(item, Location):
            return False
        return (    self.first  == item.first
                and self.last   == item.last
                and self.strand == item.strand
                and self.defect == item.defect)

    def __hash__(self):
        return hash((self._first, self._last, self._strand, self._defect))


class Feature(Copyable):
    """
    This class represents a single sequence feature, for example from a
    GenBank feature table.
    A feature describes a functional part of a sequence.
    It consists of a feature key, describing the general class of the
    feature, at least one location, describing its position on the
    reference, and qualifiers, describing the feature in detail.

    Objects of this class are immutable.

    Parameters
    ----------
    key : str
        The name of the feature class, e.g. *gene*, *CDS* or
        *regulatory*.
    locs : iterable object of Location
        A list of feature locations. In most cases this list will only
        contain one location, but multiple ones are also possible for
        example in eukaryotic CDS (due to splicing).
    qual : dict, optional
        Maps feature qualifiers to their corresponding values.
        The keys are always strings. A value is either a string or
        ``None`` if the qualifier key do not has a value.
        If key has multiple values, the values are separated by a
        line break.

    Attributes
    ----------
    key : str
        The name of the feature class, e.g. *gene*, *CDS* or
        *regulatory*.
    locs : iterable object of Location
        A list of feature locations. In most cases this list will only
        contain one location, but multiple ones are also possible for
        example in eukaryotic CDS (due to splicing).
    qual : dict
        Maps feature qualifiers to their corresponding values.
        The keys are always strings. A value is either a string or
        ``None`` if the qualifier key do not has a value.
        If key has multiple values, the values are separated by a
        line break.
    """

    def __init__(self, key, locs, qual=None):
        self._key = key
        if len(locs) == 0:
            raise ValueError("A feature must have at least one location")
        self._locs = frozenset(locs)
        self._qual = copy.deepcopy(qual) if qual is not None else {}

    def __repr__(self):
        """Represent Feature as a string for debugging."""
        return f'Feature("{self._key}", [{", ".join([loc.__repr__() for loc in self.locs])}], qual={self._qual})'

    def get_location_range(self):
        """
        Get the minimum first base/residue and maximum last base/residue
        of all feature locations.

        This can be used to create a location, that spans all of the
        feature's locations.

        Returns
        -------
        first : int
            The minimum first base/residue of all locations.
        last : int
            The maximum last base/residue of all locations.
        """
        first = np.min([loc.first for loc in self._locs])
        last = np.max([loc.last for loc in self._locs])
        return first, last

    def __eq__(self, item):
        if not isinstance(item, Feature):
            return False
        return (    self._key  == item._key
                and self._locs == item._locs
                and self._qual == item._qual)

    def __lt__(self, item):
        if not isinstance(item, Feature):
            return False
        first, last = self.get_location_range()
        it_first, it_last = item.get_location_range()
        # The first base/residue is most significant,
        # if it is equal for both features, look at last base/residue
        if first < it_first:
            return True
        elif first > it_first:
            return False
        else: # First is equal
            return last > it_last

    def __gt__(self, item):
        if not isinstance(item, Feature):
            return False
        first, last = self.get_location_range()
        it_first, it_last = item.get_location_range()
        # The first base/residue is most significant,
        # if it is equal for both features, look at last base/residue
        if first > it_first:
            return True
        elif first < it_first:
            return False
        else: # First is equal
            return last < it_last

    @property
    def key(self):
        return self._key

    @property
    def locs(self):
        return copy.copy(self._locs)

    @property
    def qual(self):
        return copy.copy(self._qual)

    def __hash__(self):
        return hash(( self._key, self._locs, frozenset(self._qual.items()) ))


class Annotation(Copyable):
    """
    An :class:`Annotation` is a set of features belonging to one
    sequence.

    Its advantage over a simple list is the base/residue position based
    indexing:
    When using slice indices in Annotation objects, a subannotation is
    created, containing copies of all :class:`Feature` objects whose
    first and last base/residue are in range of the slice.
    If the slice starts after the first base/residue or/and the slice
    ends before the last residue, the position out of range is set to
    the boundaries of the slice (the :class:`Feature` is truncated).
    In this case the :class:`Feature` obtains the
    :attr:`Location.Defect.MISS_LEFT` and/or
    :attr:`Location.Defect.MISS_RIGHT` defect.
    The third case occurs when a :class:`Feature` starts after the slice
    ends or a :class:`Feature` ends before the slice starts.
    In this case the :class:`Feature` will not appear in the
    subannotation.

    The start or stop position in the slice indices can be omitted, then
    the subannotation will include all features from the start or up to
    the stop, respectively. Step values are ignored.
    The stop values are still exclusive, i.e. the subannotation will
    contain a not truncated :class:`Feature` only if its last
    base/residue is smaller than the stop value of the slice.

    Integers or other index types are not supported. If you want to
    obtain the :class:`Feature` instances from the :class:`Annotation`
    you need to  iterate over it.
    The iteration has no defined order.
    Alternatively, you can obtain a copy of the internal
    :class:`Feature` set via :func:`get_features()`.

    Multiple :class:`Annotation` objects can be concatenated to one
    :class:`Annotation` object using the '+' operator.
    Single :class:`Feature` instances can be added this way, too.
    If a feature is present in both :class:`Annotation` objects, the
    resulting :class:`Annotation` will contain this feature twice.

    Parameters
    ----------
    features : iterable object of Feature, optional
        The features to create the :class:`Annotation` from. if not
        provided, an empty :class:`Annotation` is created.

    Examples
    --------
    Creating an annotation from a feature list:

    >>> feature1 = Feature("CDS", [Location(-10, 30 )], qual={"gene" : "test1"})
    >>> feature2 = Feature("CDS", [Location(20,  50 )], qual={"gene" : "test2"})
    >>> annotation = Annotation([feature1, feature2])
    >>> for f in sorted(list(annotation)):
    ...     print(f.qual["gene"], "".join([str(loc) for loc in f.locs]))
    test1 -10-30 >
    test2 20-50 >

    Merging two annotations and a feature:

    >>> feature3 = Feature("CDS", [Location(100, 130 )], qual={"gene" : "test3"})
    >>> feature4 = Feature("CDS", [Location(150, 250 )], qual={"gene" : "test4"})
    >>> annotation2 = Annotation([feature3, feature4])
    >>> feature5 = Feature("CDS", [Location(-50, 200 )], qual={"gene" : "test5"})
    >>> annotation = annotation + annotation2 + feature5
    >>> for f in sorted(list(annotation)):
    ...     print(f.qual["gene"], "".join([str(loc) for loc in f.locs]))
    test5 -50-200 >
    test1 -10-30 >
    test2 20-50 >
    test3 100-130 >
    test4 150-250 >

    Location based indexing, note the defects:

    >>> annotation = annotation[40:150]
    >>> for f in sorted(list(annotation)):
    ...     gene = f.qual["gene"]
    ...     loc_str = "".join([f"{loc}    {loc.defect}" for loc in f.locs])
    ...     print(gene, loc_str)
    test5 40-149 >    Defect.MISS_RIGHT|MISS_LEFT
    test2 40-50 >    Defect.MISS_LEFT
    test3 100-130 >    Defect.NONE
    """

    def __init__(self, features=None):
        if features is None:
            self._features = set()
        else:
            self._features = set(features)

    def __repr__(self):
        """Represent Annotation as a string for debugging."""
        return f'Annotation([{", ".join([feat.__repr__() for feat in self._features])}])'

    def __copy_create__(self):
        return Annotation(self._features)

    def get_features(self):
        """
        Get a copy of the internal feature set.

        Returns
        -------
        feature_list : list of Feature
            A copy of the internal feature set.
        """
        return copy.copy(self._features)

    def add_feature(self, feature):
        """
        Add a feature to the annotation.

        Parameters
        ----------
        feature : Feature
            Feature to be added.
        """
        if not isinstance(feature, Feature):
            raise TypeError(
                f"Only 'Feature' objects are supported, "
                f"not {type(feature).__name__}"
            )
        self._features.add(feature)

    def get_location_range(self):
        """
        Get the range of feature locations,
        i.e. the first and exclusive last base/residue.

        Returns
        -------
        int : start
            Start location.
        int : stop
            Exclusive stop location.
        """
        first = sys.maxsize
        last = -sys.maxsize
        for feature in self._features:
            for loc in feature.locs:
                if loc.first < first:
                    first = loc.first
                if loc.last > last:
                    last = loc.last
        # Exclusive stop -> +1
        return first, last+1

    def del_feature(self, feature):
        """
        Delete a feature from the annotation.

        Parameters
        ----------
        feature : Feature
            Feature to be removed.

        Raises
        ------
        KeyError
            If the feature is not in the annotation
        """
        self._features.remove(feature)

    def __add__(self, item):
        if isinstance(item, Annotation):
            return Annotation(self._features | item._features)
        elif isinstance(item, Feature):
            return Annotation(self._features | set([item]))
        else:
            raise TypeError(
                f"Only 'Feature' and 'Annotation' objects are supported, "
                f"not {type(item).__name__}"
            )

    def __iadd__(self, item):
        if isinstance(item, Annotation):
            self._features |= item._features
        elif isinstance(item, Feature):
            self._features.add(item)
        else:
            raise TypeError(
                f"Only 'Feature' and 'Annotation' objects are supported, "
                f"not {type(item).__name__}"
            )
        return self

    def __getitem__(self, index):
        if isinstance(index, slice):
            i_first = index.start
            # If no start or stop index is given, include all
            if i_first is None:
                i_first = -sys.maxsize
            i_last = index.stop -1
            if i_last is None:
                i_last = sys.maxsize
            sub_annot = Annotation()
            for feature in self:
                locs_in_scope = []
                for loc in feature.locs:
                    # Always true for maxsize values
                    # in case no start or stop index is given
                    if loc.first <= i_last and loc.last >= i_first:
                        # The location is at least partly in the
                        # given location range
                        # Handle defects
                        first = loc.first
                        last = loc.last
                        defect = loc.defect
                        if loc.first < i_first:
                            defect |= Location.Defect.MISS_LEFT
                            first = i_first
                        if loc.last > i_last:
                            defect |= Location.Defect.MISS_RIGHT
                            last = i_last
                        locs_in_scope.append(Location(
                            first, last, loc.strand, defect
                        ))
                if len(locs_in_scope) > 0:
                    # The feature is present in the new annotation
                    # if any of the original locations is in the new
                    # scope
                    new_feature = Feature(
                        key=feature.key, locs=locs_in_scope, qual=feature.qual
                    )
                    sub_annot.add_feature(new_feature)
            return sub_annot
        else:
            raise TypeError(
                f"'{type(index).__name__}' instances are invalid indices"
            )

    def __delitem__(self, item):
        if not isinstance(item, Feature):
            raise TypeError(
                f"Only 'Feature' objects are supported, "
                f"not {type(item).__name__}"
            )
        self.del_feature(item)

    def __iter__(self):
        return self._features.__iter__()

    def __contains__(self, item):
        return item in self._features

    def __eq__(self, item):
        if not isinstance(item, Annotation):
            return False
        return self._features == item._features

    def __len__(self):
        return len(self._features)


class AnnotatedSequence(Copyable):
    """
    An :class:`AnnotatedSequence` is a combination of a
    :class:`Sequence` and an :class:`Annotation`.

    Indexing an :class:`AnnotatedSequence` with a slice returns another
    :class:`AnnotatedSequence` with the corresponding subannotation and
    a sequence start corrected subsequence, i.e. indexing starts at 1
    with the default sequence start 1.
    The sequence start in the newly created :class:`AnnotatedSequence`
    is the start of the slice.
    Furthermore, integer indices are allowed in which case the
    corresponding symbol of the sequence is returned (also sequence
    start corrected).
    In both cases the index must be in range of the sequence, e.g. if
    sequence start is 1, index 0 is not allowed.
    Negative indices do not mean indexing from the end of the sequence,
    in contrast to the behavior in :class:`Sequence` objects.
    Both index types can also be used to modify the sequence.

    Another option is indexing with a :class:`Feature` (preferably from the
    :class:`Annotation` in the same :class:`AnnotatedSequence`).
    In this case a sequence, described by the location(s) of the
    :class:`Feature`, is returned.
    When using a :class:`Feature` for setting an
    :class:`AnnotatedSequence` with a sequence, the new sequence is
    replacing the locations of the
    :class:`Feature`.
    Note the the replacing sequence must have the same length as the
    sequence of the :class:`Feature` index.

    Parameters
    ----------
    sequence : Sequence
        The sequence.
        Usually a :class:`NucleotideSequence` or
        :class:`ProteinSequence`.
    annotation : Annotation
        The annotation corresponding to `sequence`.
    sequence_start : int, optional
        By default, the first symbol of the sequence is corresponding
        to location 1 of the features in the annotation. The location
        of the first symbol can be changed by setting this parameter.
        Negative values are not supported yet.

    Attributes
    ----------
    sequence : Sequence
        The represented sequence.
    annotation : Annotation
        The annotation corresponding to `sequence`.
    sequence_start : int
        The location of the first symbol in the sequence.

    See also
    --------
    Annotation, Sequence

    Examples
    --------
    Creating an annotated sequence

    >>> sequence = NucleotideSequence("ATGGCGTACGATTAGAAAAAAA")
    >>> feature1 = Feature("misc_feature", [Location(1,2), Location(11,12)],
    ...                    {"note" : "walker"})
    >>> feature2 = Feature("misc_feature", [Location(16,22)], {"note" : "poly-A"})
    >>> annotation = Annotation([feature1, feature2])
    >>> annot_seq = AnnotatedSequence(annotation, sequence)
    >>> print(annot_seq.sequence)
    ATGGCGTACGATTAGAAAAAAA
    >>> for f in sorted(list(annot_seq.annotation)):
    ...     print(f.qual["note"])
    walker
    poly-A

    Indexing with integers, note the sequence start correction

    >>> print(annot_seq[2])
    T
    >>> print(annot_seq.sequence[2])
    G

    indexing with slices

    >>> annot_seq2 = annot_seq[:16]
    >>> print(annot_seq2.sequence)
    ATGGCGTACGATTAG
    >>> for f in annot_seq2.annotation:
    ...     print(f.qual["note"])
    walker

    Indexing with features

    >>> print(annot_seq[feature1])
    ATAT
    >>> print(annot_seq[feature2])
    AAAAAAA
    >>> print(annot_seq.sequence)
    ATGGCGTACGATTAGAAAAAAA
    >>> annot_seq[feature1] = NucleotideSequence("CCCC")
    >>> print(annot_seq.sequence)
    CCGGCGTACGCCTAGAAAAAAA
    """

    def __init__(self, annotation, sequence, sequence_start=1):
        self._annotation = annotation
        self._sequence = sequence
        self._seqstart = sequence_start

    def __repr__(self):
        """Represent AnnotatedSequence as a string for debugging."""
        return f'AnnotatedSequence({self._annotation.__repr__()}, {self._sequence.__repr__()}, ' \
               f'sequence_start={self._seqstart})'

    @property
    def sequence_start(self):
        return self._seqstart

    @property
    def sequence(self):
        return self._sequence

    @property
    def annotation(self):
        return self._annotation

    def __copy_create__(self):
        return AnnotatedSequence(
            self._annotation.copy(), self._sequence.copy, self._seqstart)

    def reverse_complement(self, sequence_start=1):
        """
        Create the reverse complement of the annotated sequence.

        This method accurately converts the position and the strand of
        the annotation.
        The information on the sequence start is lost.

        Parameters
        ----------
        sequence_start : int, optional
            The location of the first symbol in the reverse complement
            sequence.

        Returns
        -------
            The reverse complement of the annotated sequence.
        """
        rev_seqstart = sequence_start

        rev_sequence = self._sequence.reverse().complement()

        seq_len = len(self._sequence)
        rev_features = []
        for feature in self._annotation:
            rev_locs = []
            for loc in feature.locs:
                # Transform location to the reverse complement strand
                # (seq_len-1) -> last sequence index
                # (loc.last-self._seqstart) -> location to index
                # ... + rev_seqstart -> index to location
                rev_loc_first \
                    = (seq_len-1) - (loc.last-self._seqstart) + rev_seqstart
                rev_loc_last \
                    = (seq_len-1) - (loc.first-self._seqstart) + rev_seqstart

                if loc.strand == Location.Strand.FORWARD:
                    rev_loc_strand = Location.Strand.REVERSE
                else:
                    rev_loc_strand = Location.Strand.FORWARD

                rev_loc_defect = Location.Defect.NONE
                if loc.defect & Location.Defect.MISS_LEFT:
                    rev_loc_defect |= Location.Defect.MISS_RIGHT
                if loc.defect & Location.Defect.MISS_RIGHT:
                    rev_loc_defect |= Location.Defect.MISS_LEFT
                if loc.defect & Location.Defect.BEYOND_RIGHT:
                    rev_loc_defect |= Location.Defect.BEYOND_LEFT
                if loc.defect & Location.Defect.BEYOND_LEFT:
                    rev_loc_defect |= Location.Defect.BEYOND_RIGHT
                if loc.defect & Location.Defect.UNK_LOC:
                    rev_loc_defect |= Location.Defect.UNK_LOC
                if loc.defect & Location.Defect.BETWEEN:
                    rev_loc_defect |= Location.Defect.BETWEEN

                rev_locs.append(Location(
                        rev_loc_first, rev_loc_last,
                        rev_loc_strand, rev_loc_defect
                ))
            rev_features.append(Feature(
                feature.key, rev_locs, feature.qual
            ))

        return AnnotatedSequence(
            Annotation(rev_features), rev_sequence, rev_seqstart
        )

    def __getitem__(self, index):
        if isinstance(index, Feature):
            # Concatenate subsequences for each location of the feature
            locs = index.locs
            if len(locs) == 0:
                raise ValueError("Feature does not contain any locations")
            # Start by creating an empty sequence
            sub_seq = self._sequence.copy(new_seq_code=np.array([]))
            # Locations need to be sorted, as otherwise the locations
            # chunks would be merged in the wrong order
            # The order depends on whether the locs are on the forward
            # or reverse strand
            strand = None
            for loc in locs:
                if loc.strand == strand:
                    pass
                elif strand is None:
                    strand = loc.strand
                else: # loc.strand != strand
                    raise ValueError(
                        "All locations of the feature must have the same "
                        "strand direction"
                    )
            if strand == Location.Strand.FORWARD:
                sorted_locs = sorted(
                    locs, key=lambda loc: loc.first
                )
            else:
                sorted_locs = sorted(
                    locs, key=lambda loc: loc.last, reverse=True
                )
            # Merge the sequences corresponding to the ordered locations
            for loc in sorted(locs, key=lambda loc: loc.first):
                slice_start = loc.first - self._seqstart
                # +1 due to exclusive stop
                slice_stop = loc.last - self._seqstart +1
                add_seq = self._sequence[slice_start:slice_stop]
                if loc.strand == Location.Strand.REVERSE:
                    add_seq = add_seq.reverse().complement()
                sub_seq += add_seq
            return sub_seq

        elif isinstance(index, slice):
            # Sequence start correction
            if index.start is None:
                seq_start = 0
            else:
                if index.start < self._seqstart:
                    raise IndexError(
                        f"The start of the index ({index.start}) is lower "
                        f"than the start of the sequence ({self._seqstart})"
                    )
                seq_start = index.start - self._seqstart
            if index.stop is None:
                index.start = len(self._sequence)
            else:
                seq_stop = index.stop - self._seqstart
            # New value for the sequence start, value is base position
            if index.start is None:
                rel_seq_start = self._seqstart
            else:
                rel_seq_start = index.start
            return AnnotatedSequence(self._annotation[index],
                                     self._sequence[seq_start:seq_stop],
                                     rel_seq_start)

        elif isinstance(index, numbers.Integral):
            return self._sequence[index - self._seqstart]

        else:
            raise TypeError(
                f"'{type(index).__name__}' instances are invalid indices"
            )

    def __setitem__(self, index, item):
        if isinstance(index, Feature):
            # Item must be sequence
            # with length equal to sum of location lengths
            sub_seq = item
            sub_seq_i = 0
            for loc in index.locs:
                slice_start = loc.first - self._seqstart
                # +1 due to exclusive stop
                slice_stop = loc.last - self._seqstart +1
                interval_size = slice_stop - slice_start
                self._sequence[slice_start:slice_stop] \
                    = sub_seq[sub_seq_i : sub_seq_i + interval_size]
                sub_seq_i += interval_size
        elif isinstance(index, slice):
            # Sequence start correction
            if index.start is None:
                seq_start = 0
            else:
                seq_start = index.start - self._seqstart
            if index.stop is None:
                index.start = len(self._sequence)
            else:
                seq_stop = index.stop - self._seqstart
            # Item is a Sequence
            self._sequence[seq_start:seq_stop] = item
        elif isinstance(index, numbers.Integral):
            # Item is a symbol
            self._sequence[index - self._seqstart] = item
        else:
            raise TypeError(
                f"'{type(index).__name__}' instances are invalid indices"
            )

    def __eq__(self, item):
        if not isinstance(item, AnnotatedSequence):
            return False
        return (    self.annotation == item.annotation
                and self.sequence   == item.sequence
                and self._seqstart  == item._seqstart)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.database.uniprot"
__author__ = "Maximilian Greil"
__all__ = ["Query", "SimpleQuery", "CompositeQuery", "search"]

import requests
import abc
from .check import assert_valid_response


_base_url = "https://www.uniprot.org/uniprot/"


class Query(metaclass=abc.ABCMeta):
    """
    Base class for a wrapper around a search term
    for the UniProt search service.
    """

    def __init__(self):
        pass

    @abc.abstractmethod
    def __str__(self):
        pass

    def __or__(self, operand):
        return CompositeQuery("OR", self, operand)

    def __and__(self, operand):
        return CompositeQuery("AND", self, operand)

    def __xor__(self, operand):
        return CompositeQuery("NOT", self, operand)


class CompositeQuery(Query):
    """
    A representation of an composite query
    for the UniProt search service.

    A composite query is a combination of two other queries,
    combined either with an 'AND', 'OR' or 'NOT' operator.

    Usually the user does not create instances of this class directly,
    but :class:`Query` instances are combined with
    ``|`` (OR), ``&`` (AND) or ``^`` (NOT).

    Parameters
    ----------
    operator: str, {"AND", "OR", "NOT"}
        The combination operator.
    queries : iterable object of SimpleQuery
        The queries to be combined.
    """

    def __init__(self, operator, query1, query2):
        super().__init__()
        self._op = operator
        self._q1 = query1
        self._q2 = query2

    def __str__(self):
        return "{:} {:} {:}".format(str(self._q1), self._op, str(self._q2))


def _check_brackets(term):
    """
    Check if term contains correct number of round brackets and square brackets.

    Parameters
    ----------
    term: str
        The search term.

    Returns
    -------
    bool
        True if term contains correct number of round brackets and square brackets, otherwise False.
    """
    square_count = 0
    round_count = 0
    for i in term:
        if i == "[":
            square_count += 1
        elif i == "]":
            square_count -= 1
        if i == "(":
            round_count += 1
        elif i == ")":
            round_count -= 1
        if square_count < 0:
            return False
        if round_count < 0:
            return False
    return square_count == 0 and round_count == 0


class SimpleQuery(Query):
    """
    A simple query for the UniProt search service without
    combination via 'AND', 'OR' or 'NOT'. A query consists of a search
    term and an optional field.

    A list of available search fields with description can be found
    `here <https://www.uniprot.org/help/query-fields>`_.

    Parameters
    ----------
    field : str
       The field to search the term in.
       The list of possible fields and the required search term
       formatting can be found
       `here <https://www.uniprot.org/help/query-fields>`_.
    term: str
       The search term.
    """

    # Field identifiers are taken from
    # https://www.uniprot.org/help/query-fields
    _fields = [
        "accession", "active", "annotation", "author", "cdantigen", "chebi", "citation", "cluster", "count", "created",
        "database", "ec", "evidence", "existence", "family", "fragment", "gene", "gene_exact", "goa", "host", "id",
        "inchikey", "inn", "interactor", "keyword", "length", "lineage", "mass", "method", "mnemonic", "modified",
        "name", "organelle", "organism", "plasmid", "proteome", "proteomecomponent", "replaces", "reviewed", "scope",
        "sequence", "sequence_modified", "strain", "taxonomy", "tissue", "web"
    ]

    def __init__(self, field, term):
        super().__init__()
        if field not in SimpleQuery._fields:
            raise ValueError(f"Unknown field identifier '{field}'")
        if not _check_brackets(term):
            raise ValueError(
                f"Query term contains illegal number of round brackets ( ) and/or square brackets [ ]"
            )
        for invalid_string in \
                ['"', "AND", "OR", "NOT", "\t", "\n"]:
            if invalid_string in term:
                raise ValueError(
                    f"Query contains illegal term {invalid_string}"
                )
        if " " in term:
            term = f'"{term}"'
        self._field = field
        self._term = term

    def __str__(self):
        return f"{self._field}:{self._term}"


def search(query, number=10):
    """
    Get all UniProt IDs that meet the given query requirements,
    via the UniProt search service.

    This function requires an internet connection.

    Parameters
    ----------
    query : Query
        The search query.
    number : int
        The maximum number of IDs that are obtained.

    Returns
    -------
    ids : list of str
        A list of strings containing all UniProt IDs
        that meet the query requirements.

    Notes
    -----
    A list of available search fields with description can be found
    `here <https://www.uniprot.org/help/query-fields>`_.

    Examples
    --------
    >>> query = SimpleQuery("accession", "P62988") & \
                SimpleQuery("reviewed", "yes")
    >>> ids = search(query)
    >>> print(sorted(ids))
    ['P0CG47', 'P0CG48', 'P62979', 'P62987']
    """

    params = {
        'query': str(query),
        'format': 'list',
        'limit': str(number)
    }
    r = requests.get(_base_url, params=params)
    content = r.text
    assert_valid_response(r.status_code)
    return content.split('\n')[:-1]



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This subpackage provides support for the the modern PDBx/mmCIF file
format. The :class:`PDBxFile` class provides dictionary-like access to
every field in PDBx/mmCIF files.
Additional utility functions allow conversion of these dictionaries to
:class:`AtomArray` and :class:`AtomArrayStack` objects and vice versa.
"""

__name__ = "biotite.structure.io.pdbx"
__author__ = "Patrick Kunzmann"

from .convert import *
from .file import *


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module allows estimation of secondary structure elements in protein
structures.
"""

__name__ = "biotite.structure"
__author__ = "Patrick Kunzmann"
__all__ = ["annotate_sse"]

import numpy as np
from .atoms import Atom, AtomArray, AtomArrayStack, coord
from .geometry import distance, angle, dihedral
from .filter import filter_amino_acids
from .error import BadStructureError


_radians_to_angle = 2*np.pi/360

_r_helix = ((89-12)*_radians_to_angle, (89+12)*_radians_to_angle)
_a_helix = ((50-20)*_radians_to_angle, (50+20)*_radians_to_angle)
_d2_helix = ((5.5-0.5), (5.5+0.5))
_d3_helix = ((5.3-0.5), (5.3+0.5))
_d4_helix = ((6.4-0.6), (6.4+0.6))

_r_strand = ((124-14)*_radians_to_angle, (124+14)*_radians_to_angle)
_a_strand = ((-180)*_radians_to_angle, (-125)*_radians_to_angle,
             (145)*_radians_to_angle, (180)*_radians_to_angle)
_d2_strand = ((6.7-0.6), (6.7+0.6))
_d3_strand = ((9.9-0.9), (9.9+0.9))
_d4_strand = ((12.4-1.1), (12.4+1.1))


def annotate_sse(atom_array, chain_id):
    r"""
    Calculate the secondary structure elements (SSE) of a
    peptide chain based on the `P-SEA` algorithm.
    :footcite:`Labesse1997`
    
    The annotation is based CA coordinates only, specifically
    distances and dihedral angles.
    
    Parameters
    ----------
    atom_array : AtomArray
        The atom array to annotate for.
    chain_id : str
        The chain ID to annotate for.
    
    Returns
    -------
    sse : ndarray
        An array containing the secondary structure elements,
        where the index corresponds to the index of the CA-filtered
        `atom_array`. 'a' means :math:`{\alpha}`-helix, 'b' means
        :math:`{\beta}`-strand/sheet, 'c' means coil.
    
    Notes
    -----
    Although this function is based on the original `P-SEA` algorithm,
    there are deviatons compared to the official `P-SEA` software in
    some cases. Do not rely on getting the exact same results.
    
    References
    ----------

    .. footbibliography::
    
    Examples
    --------
    
    SSE of PDB 1L2Y:
    
    >>> sse = annotate_sse(atom_array, "A")
    >>> print(sse)
    ['c' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c'
     'c' 'c']
    
    """
    # Filter all CA atoms in the relevant chain.
    ca_coord = atom_array[filter_amino_acids(atom_array) &
                          (atom_array.atom_name == "CA") &
                          (atom_array.chain_id == chain_id)].coord
    
    d2i_coord = np.full(( len(ca_coord), 2, 3 ), np.nan)
    d3i_coord = np.full(( len(ca_coord), 2, 3 ), np.nan)
    d4i_coord = np.full(( len(ca_coord), 2, 3 ), np.nan)
    ri_coord = np.full(( len(ca_coord), 3, 3 ), np.nan)
    ai_coord = np.full(( len(ca_coord), 4, 3 ), np.nan)
    
    # The distances and angles are not defined for the entire interval,
    # therefore the indices do not have the full range
    # Values that are not defined are NaN
    for i in range(1, len(ca_coord)-1):
        d2i_coord[i] = (ca_coord[i-1], ca_coord[i+1])
    for i in range(1, len(ca_coord)-2):
        d3i_coord[i] = (ca_coord[i-1], ca_coord[i+2])
    for i in range(1, len(ca_coord)-3):
        d4i_coord[i] = (ca_coord[i-1], ca_coord[i+3])
    for i in range(1, len(ca_coord)-1):
        ri_coord[i] = (ca_coord[i-1], ca_coord[i], ca_coord[i+1])
    for i in range(1, len(ca_coord)-2):
        ai_coord[i] = (ca_coord[i-1], ca_coord[i],
                       ca_coord[i+1], ca_coord[i+2])
    
    d2i = distance(d2i_coord[:,0], d2i_coord[:,1])
    d3i = distance(d3i_coord[:,0], d3i_coord[:,1])
    d4i = distance(d4i_coord[:,0], d4i_coord[:,1])
    ri = angle(ri_coord[:,0], ri_coord[:,1], ri_coord[:,2])
    ai = dihedral(ai_coord[:,0], ai_coord[:,1],
                  ai_coord[:,2], ai_coord[:,3])
    
    sse = np.full(len(ca_coord), "c", dtype="U1")
    
    # Annotate helices
    # Find CA that meet criteria for potential helices
    is_pot_helix = np.zeros(len(sse), dtype=bool)
    for i in range(len(sse)):
        if (
                d3i[i] >= _d3_helix[0] and d3i[i] <= _d3_helix[1]
            and d4i[i] >= _d4_helix[0] and d4i[i] <= _d4_helix[1]
           ) or (
                ri[i] >= _r_helix[0] and ri[i] <= _r_helix[1]
            and ai[i] >= _a_helix[0] and ai[i] <= _a_helix[1]
           ):
                is_pot_helix[i] = True
    # Real helices are 5 consecutive helix elements
    is_helix = np.zeros(len(sse), dtype=bool)
    counter = 0
    for i in range(len(sse)):
        if is_pot_helix[i]:
            counter += 1
        else:
            if counter >= 5:
                is_helix[i-counter : i] = True
            counter = 0
    # Extend the helices by one at each end if CA meets extension criteria
    i = 0
    while i < len(sse):
        if is_helix[i]:
            sse[i] = "a"
            if (
                d3i[i-1] >= _d3_helix[0] and d3i[i-1] <= _d3_helix[1]
               ) or (
                ri[i-1] >= _r_helix[0] and ri[i-1] <= _r_helix[1]
               ):
                    sse[i-1] = "a"
            sse[i] = "a"
            if (
                d3i[i+1] >= _d3_helix[0] and d3i[i+1] <= _d3_helix[1]
               ) or (
                ri[i+1] >= _r_helix[0] and ri[i+1] <= _r_helix[1]
               ):
                    sse[i+1] = "a"
        i += 1
    
    # Annotate sheets
    # Find CA that meet criteria for potential strands
    is_pot_strand = np.zeros(len(sse), dtype=bool)
    for i in range(len(sse)):
        if (    d2i[i] >= _d2_strand[0] and d2i[i] <= _d2_strand[1]
            and d3i[i] >= _d3_strand[0] and d3i[i] <= _d3_strand[1]
            and d4i[i] >= _d4_strand[0] and d4i[i] <= _d4_strand[1]
           ) or (
                ri[i] >= _r_strand[0] and ri[i] <= _r_strand[1]
            and (   (ai[i] >= _a_strand[0] and ai[i] <= _a_strand[1])
                 or (ai[i] >= _a_strand[2] and ai[i] <= _a_strand[3]))
           ):
                is_pot_strand[i] = True
    # Real strands are 5 consecutive strand elements,
    # or shorter fragments of at least 3 consecutive strand residues,
    # if they are in hydrogen bond proximity to 5 other residues
    pot_strand_coord = ca_coord[is_pot_strand]
    is_strand = np.zeros(len(sse), dtype=bool)
    counter = 0
    contacts = 0
    for i in range(len(sse)):
        if is_pot_strand[i]:
            counter += 1
            coord = ca_coord[i]
            for strand_coord in ca_coord:
                dist = distance(coord, strand_coord)
                if dist >= 4.2 and dist <= 5.2:
                    contacts += 1
        else:
            if counter >= 4:
                is_strand[i-counter : i] = True
            elif counter == 3 and contacts >= 5:
                is_strand[i-counter : i] = True
            counter = 0
            contacts = 0
    # Extend the strands by one at each end if CA meets extension criteria
    i = 0
    while i < len(sse):
        if is_strand[i]:
            sse[i] = "b"
            if d3i[i-1] >= _d3_strand[0] and d3i[i-1] <= _d3_strand[1]:
                sse[i-1] = "b"
            sse[i] = "b"
            if d3i[i+1] >= _d3_strand[0] and d3i[i+1] <= _d3_strand[1]:
                sse[i+1] = "b"
        i += 1
    
    return sse
            



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.io.gro"
__author__ = "Daniel Bauer, Patrick Kunzmann"
__all__ = ["GROFile"]

import numpy as np
from ...atoms import AtomArray, AtomArrayStack
from ...box import is_orthogonal
from ....file import TextFile, InvalidFileError
from ..general import _guess_element as guess_element
from ...error import BadStructureError
import copy
from datetime import datetime

_atom_records = {"res_id"    : (0, 5),
                 "res_name"  : (5,10),
                 "atom_name" : (10,15),
                 "atom_id"   : (15,20),
                 "coord_x"   : (20, 28),
                 "coord_y"   : (28, 36),
                 "coord_z"   : (36, 44),
                 "v_x"       : (44, 52),
                 "v_y"       : (52, 60),
                 "v_z"       : (60, 68)}


class GROFile(TextFile):
    r"""
    This class represents a GRO file.

    This class only provides support for reading/writing the pure atom
    information

    Examples
    --------
    Load a `\\*.gro` file, modify the structure and save the new
    structure into a new file:
    
    >>> import os.path
    >>> file = GROFile.read(os.path.join(path_to_structures, "1l2y.gro"))
    >>> array_stack = file.get_structure()
    >>> array_stack_mod = rotate(array_stack, [1,2,3])
    >>> file = GROFile()
    >>> file.set_structure(array_stack_mod)
    >>> file.write(os.path.join(path_to_directory, "1l2y_mod.gro"))
    
    """
    def get_model_count(self):
        """
        Get the number of models contained in this GRO file.

        Returns
        -------
        model_count : int
            The number of models.
        """
        model_count = 0
        for line in self.lines:
            if _is_int(line):
                model_count += 1
        return model_count


    def get_structure(self, model=None):
        """
        Get an :class:`AtomArray` or :class:`AtomArrayStack` from the
        GRO file.
        
        Parameters
        ----------
        model : int, optional
            If this parameter is given, the function will return an
            :class:`AtomArray` from the atoms corresponding to the given
            model number (starting at 1).
            Negative values are used to index models starting from the
            last model insted of the first model.
            If this parameter is omitted, an :class:`AtomArrayStack`
            containing all models will be returned, even if the
            structure contains only one model.
        
        Returns
        -------
        array : AtomArray or AtomArrayStack
            The return type depends on the `model` parameter.
        """
        
        def get_atom_line_i(model_start_i, model_atom_counts):
            """
            Helper function to get the indices of all atoms for a model
            """
            return np.arange(
                model_start_i+1, model_start_i+1+model_atom_counts
            )
        
        def set_box_dimen(box_param):
            """
            Helper function to create the box vectors from the values
            in the GRO file

            Parameters
            ----------
            box_param : list of float
                The box dimensions in the GRO file.
            
            Returns
            -------
            box_vectors : ndarray, dtype=float, shape=(3,3)
                The atom array compatible box vectors.
            """
            if not any(box_param):
                return None
            if len(box_param) == 3:
                x, y, z = box_param
                return np.array([[x,0,0], [0,y,0], [0,0,z]], dtype=float)
            elif len(box_param) == 9:
                x1, y2, z3, x2, x3, y1, y3, z1, z2 = box_param
                return np.array(
                    [[x1,x2,x3], [y1,y2,y3], [z1,z2,z3]], dtype=float
                )
            else:
                raise InvalidFileError(
                    f"Invalid amount of box parameters: {len(box_param)}"
                )

        # Line indices where a new model starts
        model_start_i = np.array([i for i in range(len(self.lines))
                                  if _is_int(self.lines[i])],
                                 dtype=int)

        # Number of atoms in each model
        model_atom_counts = np.array(
            [int(self.lines[i]) for i in model_start_i]
        )

        if model is None:
            # Check if all models have the same length
            if np.all(model_atom_counts != model_atom_counts[0]):
                raise BadStructureError("The models in the file have unequal "
                                        "amount of atoms, give an explicit "
                                        "model instead")
            depth = len(model_start_i)
            length = model_atom_counts[0]
            array = AtomArrayStack(depth, length)

            # Line indices for annotation determination is determined
            # from model 1
            annot_i = get_atom_line_i(model_start_i[0], length)
        else:
            if model == 0:
                raise ValueError("The model index must not be 0")
            # Negative models mean index starting from last model
            model = len(model_start_i) + model + 1 if model < 0 else model
            if model > len(model_start_i):
                raise ValueError(
                    f"The file has {len(model_start_i)} models, "
                    f"the given model {model} does not exist"
                )

            length = model_atom_counts[model-1]
            array = AtomArray(length)

            annot_i = get_atom_line_i(model_start_i[model-1], length)

        # Replace empty strings for elements with guessed types
        # i is index in array, line_i is line index
        for i, line_i in enumerate(annot_i):
            line = self.lines[line_i]
            array.res_id[i] = int(line[0:5])
            array.res_name[i] = line[5:10].strip()
            array.atom_name[i] = line[10:15].strip()
            array.element[i] = guess_element(line[10:15].strip())

        # Fill in coordinates and boxes
        if isinstance(array, AtomArray):
            atom_i = annot_i
            for i, line_i in enumerate(atom_i):
                line = self.lines[line_i]
                # gro files use nm instead of A
                array.coord[i,0] = float(line[20:28])*10
                array.coord[i,1] = float(line[28:36])*10
                array.coord[i,2] = float(line[36:44])*10
            # Box is stored in last line (after coordinates)
            box_i = atom_i[-1] + 1
            box_param = [float(e)*10 for e in self.lines[box_i].split()]
            array.box = set_box_dimen(box_param)
        
        elif isinstance(array, AtomArrayStack):
            for m in range(len(model_start_i)):
                atom_i = get_atom_line_i(
                    model_start_i[m], model_atom_counts[m]
                )
                for i, line_i in enumerate(atom_i):
                    line = self.lines[line_i]
                    array.coord[m,i,0] = float(line[20:28])*10
                    array.coord[m,i,1] = float(line[28:36])*10
                    array.coord[m,i,2] = float(line[36:44])*10
                # Box is stored in last line (after coordinates)
                box_i = atom_i[-1] + 1
                box_param = [float(e)*10 for e in self.lines[box_i].split()]
                box = set_box_dimen(box_param)
                # Create a box in the stack if not already existing
                # and the box is not a dummy
                if box is not None:
                    if array.box is None: 
                        array.box = np.zeros((array.stack_depth(), 3, 3))
                    array.box[m] = box
                    
        return array

            
    def set_structure(self, array):
        """
        Set the :class:`AtomArray` or :class:`AtomArrayStack` for the
        file.
        
        Parameters
        ----------
        array : AtomArray or AtomArrayStack
            The array or stack to be saved into this file. If a stack
            is given, each array in the stack is saved as separate
            model.
        """
        def get_box_dimen(array):
            """
            GRO files have the box dimensions as last line for each
            model.
            In case, the `box` attribute of the atom array is
            `None`, we simply use the min and max coordinates in xyz
            to get the correct size

            Parameters
            ----------
            array : AtomArray
                The atom array to get the box dimensions from.
            
            Returns
            -------
            box : str
                The box, properly formatted for GRO files.
            """
            if array.box is None:
                coord = array.coord
                bx, by, bz = (coord.max(axis=0) - coord.min(axis=0)) / 10
                return f"{bx:>8.3f} {by:>8.3f} {bz:>8.3f}"
            else:
                box = array.box
                if is_orthogonal(box):
                    bx, by, bz = np.diag(box) / 10
                    return f"{bx:>9.5f} {by:>9.5f} {bz:>9.5f}"
                else:
                    box = box / 10
                    box_elements = (
                        box[0,0], box[1,1], box[2,2],
                        box[0,1], box[0,2],
                        box[1,0], box[1,2],
                        box[2,0], box[2,1],
                    )
                    return " ".join([f"{e:>9.5f}" for e in box_elements])
        
        if "atom_id" in array.get_annotation_categories():
            atom_id = array.atom_id
        else:
            atom_id = np.arange(1, array.array_length() + 1)
        # Atom IDs are supported up to 99999,
        # but negative IDs are also possible
        gro_atom_id = np.where(
            atom_id > 0,
            ((atom_id - 1) % 99999) + 1,
            atom_id
        )
        # Residue IDs are supported up to 9999,
        # but negative IDs are also possible
        gro_res_id = np.where(
            array.res_id > 0,
            ((array.res_id - 1) % 99999) + 1,
            array.res_id
        )

        if isinstance(array, AtomArray):
            self.lines = [None] * (array.array_length() + 3)

            # Write header lines
            self.lines[0] = f"Generated by Biotite at {datetime.now()}"
            self.lines[1] = str(array.array_length())

            # Write atom lines
            fmt = "{:>5d}{:5s}{:>5s}{:>5d}{:>8.3f}{:>8.3f}{:>8.3f}"
            for i in range(array.array_length()):
                # gro format is in nm -> multiply coords by 10
                self.lines[i+2] = fmt.format(
                    gro_res_id[i], array.res_name[i], array.atom_name[i],
                    gro_atom_id[i], array.coord[i,0]/10, array.coord[i,1]/10,
                    array.coord[i,2]/10
                )
            # Write box lines
            self.lines[-1] = get_box_dimen(array)
        elif isinstance(array, AtomArrayStack):
            self.lines = []
            # The entire information, but the coordinates,
            # is equal for each model
            # Therefore template lines are created
            # which are afterwards applied for each model
            templines = [None] * array.array_length()
            fmt = '{:>5d}{:5s}{:>5s}{:5d}'
            for i in range(array.array_length()):
                templines[i] = fmt.format(gro_res_id[i], array.res_name[i],
                                           array.atom_name[i], gro_atom_id[i])

            for i in range(array.stack_depth()):
                self.lines.append(
                    f"Generated by Biotite at {datetime.now()}, model={i+1}"
                )
                self.lines.append(str(array.array_length()))

                # Fill in coordinates for each model
                modellines = copy.copy(templines)
                for j, line in enumerate(modellines):
                    # Insert coordinates
                    line = (line + "{:>8.3f}{:>8.3f}{:>8.3f}".format(
                                    array.coord[i,j,0]/10,
                                    array.coord[i,j,1]/10,
                                    array.coord[i,j,2]/10))
                    modellines[j] = line
                self.lines.extend(modellines)
                self.lines.append(get_box_dimen(array[i]))
        else:
            raise TypeError("An atom array or stack must be provided")
        # Add terminal newline, since PyMOL requires it
        self.lines.append("")


def _is_int(string):
    """
    Return ``True`, if the string can be parsed to an int.
    """
    try:
        int(string)
        return True
    except ValueError:
        return False


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import numpy as np
import pytest
import biotite.sequence as seq
import biotite.sequence.align as align


K = 3


@pytest.fixture
def kmer_alphabet():
    return align.KmerAlphabet(seq.ProteinSequence.alphabet, K)

@pytest.fixture
def spaced_kmer_alphabet():
    return align.KmerAlphabet(seq.ProteinSequence.alphabet, K, spacing=[0,1,2])



np.random.seed(0)
N = 10
L = 30
@pytest.mark.parametrize(
    "ref_split_kmer_code",
    # Test for single instances as input
    list(np.random.randint(len(seq.ProteinSequence.alphabet), size=(N, K))) +
    # Test for multiple instances as input
    list(np.random.randint(len(seq.ProteinSequence.alphabet), size=(N, L, K)))
)
def test_fuse_and_split(kmer_alphabet, ref_split_kmer_code):
    """
    Check if :meth:`fuse()` and its reverse counterpart :meth:`split()`
    work properly by using them back and forth on random input.
    """
    fused = kmer_alphabet.fuse(ref_split_kmer_code)
    test_split_kmer_code = kmer_alphabet.split(fused)
    
    assert test_split_kmer_code.tolist() == ref_split_kmer_code.tolist()


np.random.seed(0)
N = 10
@pytest.mark.parametrize(
    "split_kmer_code",
    np.random.randint(len(seq.ProteinSequence.alphabet), size=(N, K))
)
def test_encode_and_decode(kmer_alphabet, split_kmer_code):
    """
    Check if :meth:`encode()` and its reverse counterpart
    :meth:`decode()` work properly by using them back and forth on
    random input.
    """
    alph = seq.ProteinSequence.alphabet
    ref_kmer_symbol = alph.decode_multiple(split_kmer_code)
    kmer_code = kmer_alphabet.encode(ref_kmer_symbol)
    test_kmer_symbol = kmer_alphabet.decode(kmer_code)
    
    assert test_kmer_symbol.tolist() == ref_kmer_symbol.tolist()


def test_create_continuous_kmers(kmer_alphabet):
    """
    Test :meth:`create_kmers()` against repetitive use of
    :meth:`fuse()`, which rely on two different implementations.
    The input sequence code is randomly created.
    """
    np.random.seed(0)
    LENGTH = 100

    seq_code = np.random.randint(
        len(seq.ProteinSequence.alphabet), size=LENGTH, dtype=np.uint8
    )

    ref_kmers = [
        kmer_alphabet.fuse(seq_code[i : i + kmer_alphabet.k])
        for i in range(len(seq_code) - kmer_alphabet.k + 1)
    ]

    test_kmers = kmer_alphabet.create_kmers(seq_code)

    assert test_kmers.tolist() == ref_kmers


N = 50
@pytest.mark.parametrize("seed", range(N))
def test_create_spaced_kmers(kmer_alphabet, spaced_kmer_alphabet, seed):
    """
    Test :meth:`create_kmers()` for creating spaced *k-mers*.
    Compare results from random sequences to corresponding results from
    :meth:`create_kmers()` without spacing, by using a spacing model
    that is equivalent to non-spaced *k-mers*.
    """
    MIN_LENGTH = 10
    MAX_LENGTH = 1000
    np.random.seed(seed)
    sequence = seq.ProteinSequence()
    sequence.code = np.random.randint(
        len(sequence.alphabet),
        size=np.random.randint(MIN_LENGTH, MAX_LENGTH)
    )

    ref_kmers = kmer_alphabet.create_kmers(sequence.code)
    test_kmers = spaced_kmer_alphabet.create_kmers(sequence.code)

    assert len(test_kmers) == len(ref_kmers)
    assert test_kmers.tolist() == ref_kmers.tolist()


def test_invalid_spacing():
    """
    Check if expected exceptions are raised if an invalid spacing is
    given.
    """
    alphabet = seq.ProteinSequence.alphabet

    with pytest.raises(ValueError):
        # Not enough informative positions for given k
        align.KmerAlphabet(alphabet, 5, spacing=[0, 1, 3, 4])
    with pytest.raises(ValueError):
        # Duplicate positions
        align.KmerAlphabet(alphabet, 5, spacing=[0, 1, 1, 3, 4])
    with pytest.raises(ValueError):
        # Negative values
        align.KmerAlphabet(alphabet, 5, spacing=[-1, 1, 2, 3, 4])


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
A subpackage for downloading files from the RCSB PDB.
"""

__name__ = "biotite.database.rcsb"
__author__ = "Patrick Kunzmann"

from .download import *
from .query import *


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
Functions related to working with the simulation box or unit cell
of a structure 
"""

__name__ = "biotite.structure"
__author__ = "Patrick Kunzmann"
__all__ = ["vectors_from_unitcell", "unitcell_from_vectors", "box_volume",
           "repeat_box", "repeat_box_coord", "move_inside_box",
           "remove_pbc", "remove_pbc_from_coord",
           "coord_to_fraction", "fraction_to_coord", "is_orthogonal"]

from collections.abc import Iterable
from numbers import Integral
import numpy as np
import numpy.linalg as linalg
from .util import vector_dot
from .residues import get_residue_starts
from .atoms import AtomArray, AtomArrayStack
from .error import BadStructureError


def vectors_from_unitcell(len_a, len_b, len_c, alpha, beta, gamma):
    """
    Calculate the three vectors spanning a box from the unit cell
    lengths and angles.

    The return value of this function are the three box vectors as
    required for the :attr:`box` attribute in atom arrays and stacks.

    Parameters
    ----------
    len_a, len_b, len_c : float
        The lengths of the three box/unit cell vectors *a*, *b* and *c*.
    alpha, beta, gamma:
        The angles between the box vectors in radians.
        *alpha* is the angle between *b* and *c*,
        *beta* between *a* and *c*, *gamma* between *a* and *b*
    
    Returns
    -------
    box : ndarray, dtype=float, shape=(3,3)
        The three box vectors.
        The vector components are in the last dimension.
        The value can be directly used as :attr:`box` attribute in an
        atom array.
    
    See also
    --------
    unitcell_from_vectors
    """
    a_x = len_a
    b_x = len_b * np.cos(gamma)
    b_y = len_b * np.sin(gamma)
    c_x = len_c * np.cos(beta)
    c_y = len_c * (np.cos(alpha) - np.cos(beta)*np.cos(gamma)) / np.sin(gamma)
    c_z = np.sqrt(len_c*len_c - c_x*c_x - c_y*c_y)
    box = np.array([
        [a_x,   0,   0],
        [b_x, b_y,   0],
        [c_x, c_y, c_z]
    ], dtype=np.float32)
    
    # Fix numerical errors, as values, that are actually 0,
    # might not be calculated as such
    tol = 1e-4 * (len_a + len_b + len_c)
    box[np.abs(box) < tol] = 0
    
    return box


def unitcell_from_vectors(box):
    """
    Get the unit cell lengths and angles from box vectors.

    This is the reverse operation of :func:`vectors_from_unitcell()`.

    Parameters
    ----------
    box : ndarray, shape=(3,3)
        The box vectors
    
    Returns
    -------
    len_a, len_b, len_c : float
        The lengths of the three box/unit cell vectors *a*, *b* and *c*.
    alpha, beta, gamma : float
        The angles between the box vectors in radians.

    See also
    --------
    vectors_from_unitcell
    """
    a = box[0]
    b = box[1]
    c = box[2]
    len_a = linalg.norm(a)
    len_b = linalg.norm(b)
    len_c = linalg.norm(c)
    alpha = np.arccos(np.dot(b, c) / (len_b * len_c))
    beta  = np.arccos(np.dot(a, c) / (len_a * len_c))
    gamma = np.arccos(np.dot(a, b) / (len_a * len_b))
    return len_a, len_b, len_c, alpha, beta, gamma


def box_volume(box):
    """
    Get the volume of one ore multiple boxes.

    Parameters
    ----------
    box : ndarray, shape=(3,3) or shape=(m,3,3)
        One or multiple boxes to get the volume for.
    
    Returns
    -------
    volume : float or ndarray, shape=(m,)
    """
    # Using the triple product
    return np.abs(linalg.det(box))


def repeat_box(atoms, amount=1):
    r"""
    Repeat the atoms in a box by duplicating and placing them in
    adjacent boxes.

    The output atom array (stack) contains the original atoms (central
    box) and duplicates of them in the given amount of adjacent boxes.
    The coordinates of the duplicate atoms are translated accordingly
    by the box coordinates.

    Parameters
    ----------
    atoms : AtomArray or AtomArrayStack
        The atoms to be repeated.
        If `atoms` is a :class:`AtomArrayStack`, the atoms are repeated
        for each model, according to the box of each model.
    amount : int, optional
        The amount of boxes that are created in each direction of the
        central box.
        Hence, the total amount of boxes is
        :math:`(1 + 2 \cdot \text{amount}) ^ 3`.
        By default, one box is created in each direction, totalling in
        27 boxes.

    Returns
    -------
    repeated : AtomArray or AtomArrayStack
        The repeated atoms.
        Includes the original atoms (central box) in the beginning of
        the atom array (stack).
    indices : ndarray, dtype=int, shape=(n,3)
        Indices to the atoms in the original atom array (stack).
        Equal to
        ``numpy.tile(np.arange(atoms.array_length()), (1 + 2 * amount) ** 3)``.
    
    See also
    --------
    repeat_box_coord

    Examples
    --------

    >>> array = AtomArray(length=2)
    >>> array.coord = np.array([[1,5,3], [-1,2,5]], dtype=float)
    >>> array.box = np.array([[10,0,0], [0,10,0], [0,0,10]], dtype=float)
    >>> repeated, indices = repeat_box(array)
    >>> print(repeated.coord)
    [[  1.   5.   3.]
     [ -1.   2.   5.]
     [ -9.  -5.  -7.]
     [-11.  -8.  -5.]
     [ -9.  -5.   3.]
     [-11.  -8.   5.]
     [ -9.  -5.  13.]
     [-11.  -8.  15.]
     [ -9.   5.  -7.]
     [-11.   2.  -5.]
     [ -9.   5.   3.]
     [-11.   2.   5.]
     [ -9.   5.  13.]
     [-11.   2.  15.]
     [ -9.  15.  -7.]
     [-11.  12.  -5.]
     [ -9.  15.   3.]
     [-11.  12.   5.]
     [ -9.  15.  13.]
     [-11.  12.  15.]
     [  1.  -5.  -7.]
     [ -1.  -8.  -5.]
     [  1.  -5.   3.]
     [ -1.  -8.   5.]
     [  1.  -5.  13.]
     [ -1.  -8.  15.]
     [  1.   5.  -7.]
     [ -1.   2.  -5.]
     [  1.   5.  13.]
     [ -1.   2.  15.]
     [  1.  15.  -7.]
     [ -1.  12.  -5.]
     [  1.  15.   3.]
     [ -1.  12.   5.]
     [  1.  15.  13.]
     [ -1.  12.  15.]
     [ 11.  -5.  -7.]
     [  9.  -8.  -5.]
     [ 11.  -5.   3.]
     [  9.  -8.   5.]
     [ 11.  -5.  13.]
     [  9.  -8.  15.]
     [ 11.   5.  -7.]
     [  9.   2.  -5.]
     [ 11.   5.   3.]
     [  9.   2.   5.]
     [ 11.   5.  13.]
     [  9.   2.  15.]
     [ 11.  15.  -7.]
     [  9.  12.  -5.]
     [ 11.  15.   3.]
     [  9.  12.   5.]
     [ 11.  15.  13.]
     [  9.  12.  15.]]
    >>> print(indices)
    [0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0
     1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1]
    """
    if atoms.box is None:
        raise TypeError("Structure has no box")
    if not isinstance(amount, Integral):
        raise TypeError("The amount must be an integer")
    box_count = (1 + 2 * amount) ** 3
    if isinstance(atoms, AtomArray):
        repeat_atoms = AtomArray(
            box_count * atoms.array_length()
        )
    elif isinstance(atoms, AtomArrayStack):
        repeat_atoms = AtomArrayStack(
            atoms.stack_depth(), box_count * atoms.array_length()
        )
    else:
        raise TypeError("An atom array or stack is required")
    
    repeat_atoms.box = atoms.box.copy()
    if atoms.bonds is not None:
        repeat_bonds = atoms.bonds.copy()
        # Repeat the bonds list 'box_count' times
        for i in range(box_count-1):
            repeat_bonds += repeat_bonds
        repeat_atoms.bonds = atoms.repeat_bonds
    for category in atoms.get_annotation_categories():
        annot_array = atoms.get_annotation(category)
        # The atoms have the same annotation in all boxes
        repeat_atoms.set_annotation(category, np.tile(annot_array, box_count))
    
    repeat_coord, indices = repeat_box_coord(atoms.coord, atoms.box)
    repeat_atoms.coord = repeat_coord
    return repeat_atoms, indices


def repeat_box_coord(coord, box, amount=1):
    r"""
    Similar to :func:`repeat_box()`, repeat the coordinates in a box by
    duplicating and placing them in adjacent boxes.

    Parameters
    ----------
    coord : ndarray, dtype=float, shape=(n,3) or shape=(m,n,3)
        The coordinates to be repeated.
    box :  ndarray, dtype=float, shape=(3,3) or shape=(m,3,3)
        The reference box.
        If only one box is provided, i.e. the shape is *(3,3)*,
        the box is used for all models *m*, if the coordinates shape
        is *(m,n,3)*.
    amount : int, optional
        The amount of boxes that are created in each direction of the
        central box.
        Hence, the total amount of boxes is
        :math:`(1 + 2 \cdot \text{amount}) ^ 3`.
        By default, one box is created in each direction, totalling in
        27 boxes.

    Returns
    -------
    repeated : ndarray, dtype=float, shape=(n,3) or shape=(m,n,3)
        The repeated coordinates, with the same shape as the input
        `coord`.
        Includes the original coordinates (central box) in the beginning
        of the array.
    indices : ndarray, dtype=int, shape=(n,3)
        Indices to the coordiantes in the original array.
        Equal to
        ``numpy.tile(np.arange(coord.shape[-2]), (1 + 2 * amount) ** 3)``.
    """
    if not isinstance(amount, Integral):
        raise TypeError("The amount must be an integer")
    # List of numpy arrays for each box repeat
    coords_for_boxes = [coord]
    for i in range(-amount, amount+1):
        for j in range(-amount, amount+1):
            for k in range(-amount, amount+1):
                # Omit the central box
                if i != 0 or j != 0 or k != 0:
                    temp_coord = coord.copy()
                    # Shift coordinates to adjacent box/unit cell
                    translation_vec = np.sum(
                        box * np.array([i,j,k])[:, np.newaxis],
                        axis=-2
                    )
                    # 'newaxis' to perform same translation on all
                    # atoms for each model
                    temp_coord += translation_vec[..., np.newaxis, :]
                    coords_for_boxes.append(temp_coord)
    return (
        np.concatenate(coords_for_boxes, axis=-2),
        np.tile(np.arange(coord.shape[-2]), (1 + 2 * amount) ** 3)
    )


def move_inside_box(coord, box):
    r"""
    Move all coordinates into the given box, with the box vectors
    originating at *(0,0,0)*.

    Coordinates are outside the box, when they cannot be represented by
    a linear combination of the box vectors with scalar factors
    :math:`0 \le a_i \le 1`.
    In this case the affected coordinates are translated by the box
    vectors, so that they are inside the box.

    Parameters
    ----------
    coord : ndarray, dtype=float, shape=(n,3) or shape=(m,n,3)
        The coordinates for one or multiple models.
    box : ndarray, dtype=float, shape=(3,3) or shape=(m,3,3)
        The box(es) for one or multiple models.
        When `coord` is given for multiple models, :attr:`box` must be
        given for multiple models as well.
    
    Returns
    -------
    moved_coord : ndarray, dtype=float, shape=(n,3) or shape=(m,n,3)
        The moved coordinates.
        Has the same shape is the input `coord`.
    
    Examples
    --------
    
    >>> box = np.array([[10,0,0], [0,10,0], [0,0,10]], dtype=float)
    >>> inside_coord        = [ 1,  2,  3]
    >>> outside_coord       = [ 1, 22, 54]
    >>> other_outside_coord = [-4,  8,  6]
    >>> coord = np.stack([inside_coord, outside_coord, other_outside_coord])
    >>> print(coord)
    [[ 1  2  3]
     [ 1 22 54]
     [-4  8  6]]
    >>> moved_coord = move_inside_box(coord, box)
    >>> print(moved_coord.astype(int))
    [[1 2 3]
     [1 2 4]
     [6 8 6]]
    """
    fractions = coord_to_fraction(coord, box)
    fractions_rem = fractions % 1
    return fraction_to_coord(fractions_rem, box)


def remove_pbc(atoms, selection=None):
    """
    Remove segmentation caused by periodic boundary conditions from a
    given structure.

    In this process the first atom (of the selection) is taken as origin
    and is moved inside the box.
    All other coordinates are assembled relative to the origin.
    
    Parameters
    ----------
    atoms : AtomArray or AtomArrayStack
        The potentially segmented structure.
        The :attr:`box` attribute must be set in the structure.
    selection : str or (iterable object of) ndarray, dtype=bool, shape=(n,), optional
        Specifies which part(s) of structure are sanitized, i.e the
        segmentation is removed.
        If a string is given, the value is interpreted as the chain ID
        to be selected.
        If a boolean mask is given, the corresponding atoms are
        selected. 
        If multiple boolean masks are given, each selection
        is treated as separate assembly process, independent of all
        other selections.
        Consequently, giving multiple boolean masks has the same result
        as calling the functions multiple times with each mask
        separately.
        An atom must not be selected more than one time.

    
    Returns
    -------
    sanitized_atoms : AtomArray or AtomArrayStack
        The input structure with removed periodic boundary conditions.
    
    See also
    --------
    remove_pbc_from_coord

    Notes
    -----
    It is not recommended to select regions of the
    structure with distances from one atom to the next atom that are
    larger than half of the box size
    (e.g. the solvent, chain transitions).
    In this case, multiple selections should be given, with a single
    molecule selected in each selection.

    Internally the function uses :func:`remove_pbc_from_coord()`.
    """
    if atoms.box is None:
        raise BadStructureError(
            "The 'box' attribute must be set in the structure"
        )
    new_atoms = atoms.copy()
    
    if selection is None:
        new_atoms.coord = remove_pbc_from_coord(
            atoms.coord, atoms.box
        )

    elif isinstance(selection, str):
        # Chain ID
        selection = (atoms.chain_id == selection)
        new_atoms.coord[..., selection, :] = remove_pbc_from_coord(
            atoms.coord[..., selection, :], atoms.box
        )
    
    elif isinstance(selection, np.ndarray) and selection.ndim == 1:
        # Single boolean mask
        new_atoms.coord[..., selection, :] = remove_pbc_from_coord(
            atoms.coord[..., selection, :], atoms.box
        )
    
    elif isinstance(selection, Iterable):
        # Iterable of boolean masks
        selections = np.stack(list(selection))
        # Test whether an atom was selected multiple times
        sel_count = np.count_nonzero(selections, axis=0)
        if (sel_count > 1).any():
            first_pos = np.where((sel_count > 1))[0][0]
            raise ValueError(
                f"Atom at index {first_pos} was selected "
                f"{sel_count[first_pos]} times"
            )
        for selection in selections:
            new_atoms.coord[..., selection, :] = remove_pbc_from_coord(
                atoms.coord[..., selection, :], atoms.box
            )

    return new_atoms


def remove_pbc_from_coord(coord, box):
    """
    Remove segmentation caused by periodic boundary conditions from
    given coordinates.

    In this process the first coordinate is taken as origin and
    is moved inside the box.
    All other coordinates are assembled relative to the origin by using
    the displacement coordinates in adjacent array positions.
    Basically, this function performs the reverse action of
    :func:`move_inside_box()`.
    
    Parameters
    ----------
    coord : ndarray, dtype=float, shape=(m,n,3) or shape=(n,3)
        The coordinates of the potentially segmented structure. 
    box : ndarray, dtype=float, shape=(m,3,3) or shape=(3,3)
        The simulation box or unit cell that is used as periodic
        boundary.
        The amount of dimensions must fit the `coord` parameter.

    Returns
    -------
    sanitized_coord : ndarray, dtype=float, shape=(m,n,3) or shape=(n,3)
        The reassembled coordinates.
    
    See also
    --------
    remove_pbc_from_coord
    move_inside_box

    Notes
    -----
    This function solves a common problem from MD simulation output:
    When atoms move over the periodic boundary, they reappear on the
    other side of the box, segmenting the structure.
    This function reassembles the given coordinates, removing the
    segmentation.
    """

    # Import in function to avoid circular import
    from .geometry import index_displacement
    # Get the PBC-sanitized displacements of all coordinates
    # to the respective next coordinate
    index_pairs = np.stack(
        [
            np.arange(0, coord.shape[-2] - 1), 
            np.arange(1, coord.shape[-2]    )
        ],
        axis=1
    )
    neighbour_disp = index_displacement(
        coord, index_pairs, box=box, periodic=True
    )
    # Get the PBC-sanitized displacements of all but the first
    # coordinates to (0,0,0)
    absolute_disp = np.cumsum(neighbour_disp, axis=-2)
    # The first coordinate should be inside the box
    base_coord = move_inside_box(coord[..., 0:1, :], box)
    # The new coordinates are obtained by adding the displacements
    # to the new origin
    sanitized_coord = np.zeros(coord.shape, coord.dtype)
    sanitized_coord[..., 0:1, :] = base_coord
    sanitized_coord[..., 1:, :] = base_coord + absolute_disp
    return sanitized_coord


def coord_to_fraction(coord, box):
    """
    Transform coordinates to fractions of box vectors.

    Parameters
    ----------
    coord : ndarray, dtype=float, shape=(n,3) or shape=(m,n,3)
        The coordinates for one or multiple models.
    box : ndarray, dtype=float, shape=(3,3) or shape=(m,3,3)
        The box(es) for one or multiple models.
        When `coord` is given for multiple models, :attr:`box` must be
        given for multiple models as well.
    
    Returns
    -------
    fraction : ndarray, dtype=float, shape=(n,3) or shape=(m,n,3)
        The fractions of the box vectors.
    
    See also
    --------
    fraction_to_coord

    Examples
    --------
    
    >>> box = np.array([[5,0,0], [0,5,0], [0,5,5]], dtype=float)
    >>> coord = np.array(
    ...     [[1,1,1], [10,0,0], [0,0,10], [-5,2,1]],
    ...     dtype=float
    ... )
    >>> print(coord)
    [[ 1.  1.  1.]
     [10.  0.  0.]
     [ 0.  0. 10.]
     [-5.  2.  1.]]
    >>> fractions = coord_to_fraction(coord, box)
    >>> print(fractions)
    [[ 0.2  0.0  0.2]
     [ 2.0  0.0  0.0]
     [ 0.0 -2.0  2.0]
     [-1.0  0.2  0.2]]
    """
    return np.matmul(coord, linalg.inv(box))


def fraction_to_coord(fraction, box):
    """
    Transform fractions of box vectors to coordinates.

    This is the reverse operation of :func:`coord_to_fraction()`.

    Parameters
    ----------
    fraction : ndarray, dtype=float, shape=(n,3) or shape=(m,n,3)
        The fractions of the box vectors for one or multiple models.
    box : ndarray, dtype=float, shape=(3,3) or shape=(m,3,3)
        The box(es) for one or multiple models.
        When `coord` is given for multiple models, :attr:`box` must be
        given for multiple models as well.
    
    Returns
    -------
    coord : ndarray, dtype=float, shape=(n,3) or shape=(m,n,3)
        The coordinates.
    
    See also
    --------
    coord_to_fraction
    """
    return np.matmul(fraction, box)


def is_orthogonal(box):
    """
    Check, whether a box or multiple boxes is/are orthogonal.

    A box is orthogonal when the dot product of all box vectors with
    each other is 0.

    Parameters
    ----------
    box : ndarray, dtype=float, shape=(3,3) or shape=(m,3,3)
        A single box or multiple boxes.
    
    Returns
    -------
    is_orthgonal : bool or ndarray, shape=(m,), dtype=bool
        True, if the box vectors are orthogonal, false otherwise
    
    Notes
    -----
    Due to possible numerical errors, this function also evaluates two
    vectors as orthogonal, when their dot product is not exactly zero,
    but it is within a small tolerance (:math:`10^{-6}`).
    """
    # Fix numerical errors, as values, that are actually 0,
    # might not be calculated as such
    tol = 1e-6
    return (np.abs(vector_dot(box[..., 0, :], box[..., 1, :])) < tol) & \
           (np.abs(vector_dot(box[..., 0, :], box[..., 2, :])) < tol) & \
           (np.abs(vector_dot(box[..., 1, :], box[..., 2, :])) < tol)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
A subpackage that provides interfaces to the *ViennaRNA* software
package.

Secondary structures can be predicted using *RNAfold* and plotted using
*RNAplot*.
"""

__name__ = "biotite.application.viennarna"
__author__ = "Tom David Müller"

from .rnafold import *
from .rnaplot import *



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This subpackage is used for reading and writing sequence objects
using the popular FASTA format.

This package contains the :class:`FastaFile`, which provides a
dictionary like interface to FASTA files, where the header lines are
keys and the strings containing sequence data are the corresponding
values.

Furthermore, the package contains convenience functions for
getting/setting directly :class:`Sequence` objects, rather than strings.
"""

__name__ = "biotite.sequence.io.fasta"
__author__ = "Patrick Kunzmann"

from .file import *
from .convert import *


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module provides functionality for pseudoknot detection.
"""

__name__ = "biotite.structure"
__author__ = "Tom David Müller"
__all__ = ["pseudoknots"]

import numpy as np
import networkx as nx
from itertools import chain, product

def pseudoknots(base_pairs, scores=None, max_pseudoknot_order=None):
    """
    Identify the pseudoknot order for each base pair in a given set of
    base pairs.

    By default the algorithm removes base pairs until the remaining
    base pairs are completely nested i.e. no pseudoknots appear.
    The pseudoknot order of the removed base pairs is incremented and
    the procedure is repeated with these base pairs.
    Base pairs are removed in a way that maximizes the number of
    remaining base pairs.
    However, an optional score for each individual base pair can be
    provided.

    Parameters
    ----------
    base_pairs : ndarray, dtype=int, shape=(n,2)
        The base pairs to determine the pseudoknot order of. Each row
        represents indices form two paired bases. The structure of
        the :class:`ndarray` is equal to the structure of the output of
        :func:`base_pairs()`, where the indices represent the
        beginning of the residues.
    scores : ndarray, dtype=int, shape=(n,), optional
        The score for each base pair.
        By default, the score of each base pair is ``1``.
    max_pseudoknot_order : int, optional
        The maximum pseudoknot order to be found. If a base pair would
        be of a higher order, its order is specified as ``-1``.
        By default, the algorithm is run until all base pairs
        have an assigned pseudoknot order.

    Returns
    -------
    pseudoknot_order : ndarray, dtype=int, shape=(m,n)
        The pseudoknot order of the input `base_pairs`.
        Multiple solutions that maximize the number of basepairs or
        the given score, respectively, may be possible.
        Therefore all *m* individual solutions are returned.

    Notes
    -----
    The dynamic programming approach by Smit *et al*
    :footcite:`Smit2008` is applied to detect pseudoknots.
    The algorithm was originally developed to remove pseudoknots from a
    structure.
    However, if it is run iteratively on removed knotted pairs it can be
    used to identify the pseudoknot order.

    The pseudoknot order is defined as the minimum number of base pair
    set decompositions resulting in a nested structure
    :footcite:`Antczak2018`.
    Therefore, there are no pseudoknots between base pairs with the same
    pseudoknot order.

    Examples
    --------
    Remove the pseudoknotted base pair for the sequence *ABCbac*, where
    the corresponding big and small letters each represent a base pair:

    Define the base pairs as :class:`ndarray`:

    >>> basepairs = np.array([[0, 4],
    ...                       [1, 3],
    ...                       [2, 5]])

    Find the unknotted base pairs, optimizing for the maximum number of
    base pairs:

    >>> print(pseudoknots(basepairs, max_pseudoknot_order=0))
    [[ 0  0 -1]]

    This indicates that the base pair *Cc* is a pseudoknot.

    Given the length of the sequence (6 bases), we can also represent
    the unknotted structure in dot bracket notation:

    >>> print(dot_bracket(basepairs, 6, max_pseudoknot_order=0)[0])
    ((.)).

    If the maximum pseudoknot order is not restricted, the order of the
    knotted pairs is determined and can be represented using dot bracket
    letter notation:

    >>> print(pseudoknots(basepairs))
    [[0 0 1]]
    >>> print(dot_bracket(basepairs, 6)[0])
    (([))]

    See Also
    --------
    base_pairs
    dot_bracket

    References
    ----------

    .. footbibliography::

    """
    # List containing the results
    results = [np.full(len(base_pairs), -1, dtype='int32')]

    # if no score array is given, each base pairs' score is one
    if scores is None:
        scores = np.ones(len(base_pairs))

    # Make sure `base_pairs` has the same length as the score array
    if len(base_pairs) != len(scores):
        raise ValueError(
            "'base_pair' and 'scores' must have the same shape"
        )

    # Split the base pairs in regions
    regions = _find_regions(base_pairs, scores)

    # Compute results
    results = _get_results(regions, results, max_pseudoknot_order)

    return np.vstack(results)


class _Region():
    """
    This class represents a paired region.

    A region is a set of base pairs. This class provides methods to
    access the minimum and maximum index of the bases that are part of
    the region, handles score calculation, and backtracing to the
    original base pair array.

    Parameters
    ----------
    base_pairs: ndarray, shape=(n,2), dtype=int
        All base pairs of the structure the region is a subset for.
    region_pairs: ndarray, dtype=int
        The indices of the base pairs in ``base_pairs`` that are part of
        the region.
    scores : ndarray, dtype=int, shape=(n,) (default: None)
        The score for each base pair.
    """

    def __init__ (self, base_pairs, region_pairs, scores):
        # The Start and Stop indices for each Region
        self.start = np.min(base_pairs[region_pairs])
        self.stop = np.max(base_pairs[region_pairs])

        self.region_pairs = region_pairs
        self.score = np.sum(scores[region_pairs])

    def get_index_array(self):
        """
        Return an index array with the positions of the region`s bases
        in the original base pair array.

        Returns
        -------
        region_pairs : ndarray
            The indices of the bases in the original base pair array.
        """
        return self.region_pairs

    def __lt__(self, other):
        """
        This comparison operator is required for :func:`np.unique()`. As
        only the difference between the regions is relevant and not any
        particular order, a distinction is made by the objects unique
        ids.

        Parameters
        ----------
        other : _region
            The other region.

        Returns
        -------
        comparision : bool
            The evaluated comparison.
        """
        return id(self) < id(other)


def _find_regions(base_pairs, scores):
    """
    Find regions in a base pair array. A region is defined as a set of
    consecutively nested base pairs.

    Parameters
    ----------
    base_pairs : ndarray, dtype=int, shape=(n, 2)
        Each row is equivalent to one base pair and contains the first
        indices of the residues corresponding to each base.
    scores : ndarray, dtype=int, shape=(n,) (default: None)
        The score for each base pair.

    Returns
    -------
    regions : Graph
        The ``_Region`` objects as graph, where the edges represent
        conflicts.
    """
    # Make sure the lower residue is on the left for each row
    sorted_base_pairs = np.sort(base_pairs, axis=1)

    # Sort the first column in ascending order
    original_indices = np.argsort(sorted_base_pairs[:, 0])
    sorted_base_pairs = sorted_base_pairs[original_indices]

    # Rank each base
    # E.g.: [[3, 5]  --> [[0, 1]
    #        [9, 7]]      [3, 2]]
    order = np.argsort(sorted_base_pairs.flatten())
    rank = np.argsort(order).reshape(base_pairs.shape)

    # The base pairs belonging to the current region
    region_pairs = []
    # The individual regions
    regions = set()

    # Find separate regions
    for i in range(len(sorted_base_pairs)):
        # if a new region is to be started append the current base pair
        if len(region_pairs) == 0:
            region_pairs.append(original_indices[i])
            continue

        # Check if the current base pair belongs to the region that is
        # currently being defined
        previous_upstream_rank = rank[i-1, 0]
        this_upstream_rank = rank[i, 0]
        previous_downstream_rank = rank[i-1, 1]
        this_downstream_rank = rank[i, 1]

        # if the current base pair belongs to a new region, save the
        # current region and start a new region
        if ((previous_downstream_rank - this_downstream_rank) != 1 or
            (this_upstream_rank - previous_upstream_rank) != 1):
                regions.add(
                    _Region(base_pairs, np.array(region_pairs), scores)
                )
                region_pairs = []

        # Append the current base pair to the region
        region_pairs.append(original_indices[i])

    # The last region has no endpoint defined by the beginning of a
    # new region.
    regions.add(_Region(base_pairs, np.array(region_pairs), scores))

    # Return the graphical representation of the conflicting regions
    return _generate_graphical_representation(regions)


def _generate_graphical_representation(regions):
    """
    Find the conflicting regions and represent them graphically using
    the ``Graph`` class from ``Networkx``.

    Parameters
    ----------
    regions : set {_region, ...}
        The regions representing the consecutively nested base pairs.

    Returns
    -------
    regions : Graph
        The ``_Region`` objects as graph, where the edges represent
        conflicts.
    """

    # Create a graph
    region_graph = nx.Graph()

    # Add the regions to the graph as nodes
    region_graph.add_nodes_from(regions)

    # Get the region array and a boolean array, where the start of each
    # region is ``True``.
    region_array, (start_stops,) = _get_region_array_for(
        regions, content=[lambda a : [True, False]], dtype=['bool']
    )

    # Check each region for conflicts with other regions
    for start, region in enumerate(region_array):
        # Check each region only once
        if not start_stops[start]:
            continue

        # Find the index of the stopping of the region in the region
        # array
        stop = _get_first_occurrence_for(region_array[start+1:], region)
        stop += (start + 1)

        # Store regions the current region conflicts with
        conflicts = set()

        # Iterate over the regions between the starting and stopping
        # point of the current region
        for other_region in region_array[start+1:stop]:
            # If the other region is not already a conflict, add it to
            # the conflict set
            if other_region not in conflicts:
                conflicts.add(other_region)
            # If the other region is twice between the starting and
            # stopping point of the current region, its starting and
            # stopping point lie between the current region and it is
            # thus non-conflicting
            else:
                conflicts.remove(other_region)

        # Conflicts between regions are represented as graph edges
        edges = []

        # Convert the edges in a ``NetworkX`` compatible format
        for conflict in conflicts:
            edges.append((region, conflict))

        # Add the edges to the graph
        region_graph.add_edges_from(edges)
    return region_graph


def _get_first_occurrence_for(iterable, wanted_object):
    """
    Get the first occurrence of an object in an iterable.

    Parameters
    ----------
    iterable : iterable
        The iterable containing the object.
    wanted_object : object
        The object to be found.

    Returns
    -------
    index : int
        The index of the first occurrence of the object.
    """
    for i, value in enumerate(iterable):
        if value is wanted_object:
            return i


def _get_region_array_for(regions, content=[], dtype=[]):
    """
    Get a :class:`ndarray` of region objects. Each object occurs twice,
    representing its start and end point. The regions positions in the
    array reflect their relative positions.

    Furthermore, a list of functions can be provided enabling custom
    outputs for each objects` start and end point.

    Parameters
    ----------
    regions : set {_region, ...}
        The regions to be considered
    content : list [function, ...] (default: [])
        The functions to be considered for custom outputs. For a given
        region they must return a tuple of which the first value is
        placed at the start position and the second value at the end
        position of the region relative to the other regions.
    dtype : list [str, ...] (default: [])
        The data type of the output of the custom functions.

    Returns
    -------
    region_array : ndarray, dtype=object
        The array of ordered region objects.
    custom_content : list [ndarray, ...]
        The custom output.
    """
    # region_array and index array
    region_array = np.empty(len(regions)*2, dtype=_Region)
    index_array = np.empty(len(regions)*2, dtype='int32')

    # Content array for custom return arrays
    content_list = [None]*len(content)
    for i in range(len(content)):
        content_list[i] = np.empty(len(regions)*2, dtype=dtype[i])

    # Fill the arrays
    for i, reg in enumerate(regions):
        indices = [2*i, 2*i+1]
        region_array[indices] = reg
        for c in range(len(content_list)):
            content_list[c][indices] = content[c](reg)
        index_array[indices] = [reg.start, reg.stop]

    # Order the arrays by the base indices
    sort_mask = np.argsort(index_array)
    region_array = region_array[sort_mask]

    # if no custom array content is given only return the ordered array
    # containing the regions
    if content == []:
        return region_array

    # if custom content is given also return the ordered content
    for i in range(len(content_list)):
        content_list[i] = content_list[i][sort_mask]
    return region_array, content_list


def _remove_pseudoknots(regions):
    """
    Get the optimal solutions according to the algorithm referenced in
    :func:`pseudoknots()`.

    The algorithm uses a dynamic programming matrix in order to find
    the optimal solutions with the highest combined region scores.

    Parameters
    ----------
    regions : set {_region, ...}
        The conflicting regions for whích optimal solutions are to be
        found.
    scores : ndarray
        The score array.

    Returns
    -------
    solutions : ndarray, dtype=object
        The optimal solutions. Each solution in the ``ndarray`` is
        represented as ``set`` of unknotted regions.
    """
    # Create dynamic programming matrix
    dp_matrix_shape = len(regions)*2, len(regions)*2
    dp_matrix = np.empty(dp_matrix_shape, dtype='object')
    dp_matrix_solutions_starts = np.zeros_like(dp_matrix)
    dp_matrix_solutions_stops = np.zeros_like(dp_matrix)

    # Each index corresponds to the position in the dp matrix.
    # ``region_array`` contains the region objects and ``start_stops``
    # contains the lowest and highest positions of the regions
    region_array, (start_stops,) = _get_region_array_for(
        regions,
        [lambda a : (a.start, a.stop)],
        ['int32']
    )
    # Initialise the matrix diagonal with ndarrays of empty frozensets
    for i in range(len(dp_matrix)):
        dp_matrix[i, i] = np.array([frozenset()])

    # Iterate through the top right half of the dynamic programming
    # matrix
    for j in range(len(regions)*2):
        for i in range(j-1, -1, -1):
            solution_candidates = set()
            left = dp_matrix[i, j-1]
            bottom = dp_matrix[i+1, j]

            # Add all solutions of the cell to the left
            for solution in left:
                solution_candidates.add(solution)

            # Add all solutions of the cell to the bottom
            for solution in bottom:
               solution_candidates.add(solution)

            # Check if i and j are start/end-points of the same region
            if region_array[i] is region_array[j]:

                # Add all solutions from the cell to the bottom left
                # plus this region
                bottom_left = dp_matrix[i+1, j-1]
                for solution in bottom_left:
                    solution_candidates.add(solution | set([region_array[i]]))

            # Perform additional tests if solution in the left cell and
            # bottom cell both differ from an empty solution
            if (np.any(left != [frozenset()]) and
                np.any(bottom != [frozenset()])):

                left_highest = dp_matrix_solutions_stops[i, j-1]
                bottom_lowest = dp_matrix_solutions_starts[i+1, j]

                # For each pair of solutions check if solutions are
                # disjoint
                for solution1, highest in zip(left, left_highest):
                    for solution2, lowest in zip(bottom, bottom_lowest):
                        if highest < lowest:
                            # Both solutions are disjoint
                            solution_candidates.add(solution1 | solution2)
                        else:
                            # Both solutions are not disjoint
                            # Add subsolutions
                            for k in range(
                                np.where(start_stops==lowest)[0][0]-1,
                                np.where(start_stops==highest)[0][0]+1
                            ):
                                cell1 = dp_matrix[i, k]
                                cell2 = dp_matrix[k+1, j]
                                for subsolution1 in cell1:
                                    for subsolution2 in cell2:
                                        solution_candidates.add(
                                            subsolution1 | subsolution2
                                        )

            # Make solution candidates ``ndarray`` array of sets
            solution_candidates = np.array(list(solution_candidates))

            # Calculate the scores for each solution
            solution_scores = np.zeros(len(solution_candidates))
            for s, solution in enumerate(solution_candidates):
                score = 0
                for reg in solution:
                    score += reg.score
                solution_scores[s] = score
            # Get the indices where the score is at a maximum
            highest_scores = np.argwhere(
                solution_scores == np.amax(solution_scores)
            ).flatten()

            # Get the solutions with the highest score
            solution_candidates = solution_candidates[highest_scores]

            # Add the solutions to the dynamic programming matrix
            dp_matrix[i, j] = solution_candidates

            solution_starts = np.zeros_like(solution_candidates, dtype='int32')
            solution_stops = np.zeros_like(solution_candidates, dtype='int32')

            for s, solution in enumerate(solution_candidates):
                solution_starts[s] = min(
                    [reg.start for reg in solution], default=-1
                )
                solution_stops[s] = max(
                    [reg.stop for reg in solution], default=-1
                )

            dp_matrix_solutions_starts[i, j] = solution_starts
            dp_matrix_solutions_stops[i, j] = solution_stops

    # The top right corner contains the optimal solutions
    return dp_matrix[0, -1]


def _get_results(regions, results, max_pseudoknot_order, order=0):
    """
    Use the dynamic programming algorithm to get the pseudoknot order
    of a given set of regions. If there are remaining conflicts their
    results are recursively calculated and merged with the current
    results.

    Parameters
    ----------
    regions : set {_region, ...}
        The regions for whích optimal solutions are to be found.
    results : list [ndarray, ...]
        The results
    max_pseudoknot_order : int
        The maximum pseudoknot order to be found. If a base pair would
        be of a higher order, its order is specified as -1. If ``None``
        is given, all base pairs are evaluated.
    order : int (default: 0)
        The order that is currently evaluated.

    Returns
    -------
    results : list [ndarray, ...]
        The results
    """

    # Remove non-conflicting regions
    non_conflicting = [isolate for isolate in nx.isolates(regions)]
    regions.remove_nodes_from(non_conflicting)

    # Non-conflicting regions are of the current order:
    index_list_non_conflicting = list(
            chain(
                *[region.get_index_array() for region in non_conflicting]
            )
        )
    for result in results:
        result[index_list_non_conflicting] = order


    # If no conflicts remain, the results are complete
    if len(regions) == 0:
        return results

    # Get the optimal solutions for given regions. Evaluate each clique
    # of mutually conflicting regions seperately
    cliques = [component for component in nx.connected_components(regions)]
    solutions = [set(chain(*e)) for e in product(
        *[_remove_pseudoknots(clique) for clique in cliques]
    )]

    # Get a copy of the current results for each optimal solution
    results_list = [
        [result.copy() for result in results] for _ in range(len(solutions))
    ]

    # Evaluate each optimal solution
    for i, solution in enumerate(solutions):

        # Get the pseudoknotted regions
        pseudoknotted_regions = regions.copy()
        pseudoknotted_regions.remove_nodes_from(solution)

        # Get an index list of the unknotted base pairs
        index_list_unknotted = list(
            chain(
                *[region.get_index_array() for region in solution]
            )
        )

        # Write results for current solution
        for j, result in enumerate(results_list[i]):
            result[index_list_unknotted] = order

        # If this order is the specified maximum order, stop evaluation
        if max_pseudoknot_order == order:
            continue

        # Evaluate the pseudoknotted region
        results_list[i] = _get_results(
            pseudoknotted_regions, results_list[i],
            max_pseudoknot_order, order=order+1
        )

    # Flatten the results
    return list(chain(*results_list))



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.application"
__author__ = "Patrick Kunzmann"
__all__ = ["Application", "AppStateError", "TimeoutError", "VersionError",
           "AppState", "requires_state"]

import abc
import time
from functools import wraps
from enum import Flag, auto


class AppState(Flag):
    """
    This enum type represents the app states of an application. 
    """
    CREATED = auto()
    RUNNING = auto()
    FINISHED = auto()
    JOINED = auto()
    CANCELLED = auto()


def requires_state(app_state):
    """
    A decorator for methods of :class:`Application` subclasses that
    raises an :class:`AppStateError` in case the method is called, when
    the :class:`Application` is not in the specified :class:`AppState`
    `app_state`.
    
    Parameters
    ----------
    app_state : AppState
        The required app state.
    
    Examples
    --------
    Raises :class:`AppStateError` when `function` is called,
    if :class:`Application` is not in one of the specified states:
    
    >>> @requires_state(AppState.RUNNING | AppState.FINISHED)
    ... def function(self):
    ...     pass
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # First parameter of method is always 'self'
            instance = args[0]
            if not instance._state & app_state:
                raise AppStateError(
                    f"The application is in {instance.get_app_state()} state, "
                    f"but {app_state} state is required"
                )
            return func(*args, **kwargs)
        return wrapper
    return decorator


class Application(metaclass=abc.ABCMeta):
    """
    This class is a wrapper around an external piece of runnable
    software in any sense. Subclasses of this abstract base class
    specify the respective kind of software and the way of interacting
    with it.
    
    Every :class:`Application` runs through a different app states
    (instances of enum :class:`AppState`) from its creation until its
    termination:
    Directly after its instantiation the app is in the *CREATED* state.
    In this state further parameters can be set for the application run.
    After the user calls the :func:`start()` method, the app state is
    set to *RUNNING* and the :class:`Application` type specific
    :func:`run()` method is called.
    When the application finishes the AppState changes to *FINISHED*.
    This is checked via the :class:`Application` type specific
    :func:`is_finished()` method.
    The user can now call the :func:`join()` method, concluding the
    application in the *JOINED* state and making the results of the
    application accessible by executing the :class:`Application`
    type specific :func:`evaluate()` method.
    Furthermore this executes the :class:`Application` type specific
    :func:`clean_up()` method.
    :func:`join()` can even be called in the *RUNNING* state:
    This will constantly check :func:`is_finished()` and will directly
    go into the *JOINED* state as soon as the application reaches the
    *FINISHED* state.
    Calling the :func:`cancel()` method while the application is
    *RUNNING* or *FINISHED* leaves the application in the *CANCELLED*
    state.
    This triggers the :func:`clean_up()` method, too, but there are no
    accessible results.
    If a method is called in an unsuitable app state, an
    :class:`AppStateError` is called.
    
    The application run behaves like an additional thread: Between the
    call of :func:`start()` and :func:`join()` other Python code can be
    executed, while the application runs in the background.
    """
    
    def __init__(self):
        self._state = AppState.CREATED
    
    @requires_state(AppState.CREATED)
    def start(self):
        """
        Start the application run and set its state to *RUNNING*.
        This can only be done from the *CREATED* state.
        """
        self.run()
        self._start_time = time.time()
        self._state = AppState.RUNNING
    
    @requires_state(AppState.RUNNING | AppState.FINISHED)
    def join(self, timeout=None):
        """
        Conclude the application run and set its state to *JOINED*.
        This can only be done from the *RUNNING* or *FINISHED* state.
        
        If the application is *FINISHED* the joining process happens
        immediately, if otherwise the application is *RUNNING*, this
        method waits until the application is *FINISHED*.
        
        Parameters
        ----------
        timeout : float, optional
            If this parameter is specified, the :class:`Application`
            only waits for finishing until this value (in seconds) runs
            out.
            After this time is exceeded a :class:`TimeoutError` is
            raised and the application is cancelled.
        
        Raises
        ------
        TimeoutError
            If the joining process exceeds the `timeout` value.
        """
        time.sleep(self.wait_interval())
        while self.get_app_state() != AppState.FINISHED:
            if timeout is not None and time.time()-self._start_time > timeout:
                self.cancel()
                raise TimeoutError(
                    f"The application expired its timeout "
                    f"({timeout:.1f} s)"
                )
            else:
                time.sleep(self.wait_interval())
        time.sleep(self.wait_interval())
        try:
            self.evaluate()
        except AppStateError:
            raise
        except:
            self._state = AppState.CANCELLED
            raise
        else:
            self._state = AppState.JOINED
        self.clean_up()
    
    @requires_state(AppState.RUNNING | AppState.FINISHED)
    def cancel(self):
        """
        Cancel the application when in *RUNNING* or *FINISHED* state.
        """
        self._state = AppState.CANCELLED
        self.clean_up()
    
    def get_app_state(self):
        """
        Get the current app state.
        
        Returns
        -------
        app_state : AppState
            The current app state.
        """
        if self._state == AppState.RUNNING:
            if self.is_finished():
                self._state = AppState.FINISHED
        return self._state
    
    @abc.abstractmethod
    def run(self):
        """
        Commence the application run. Called in :func:`start()`.
        
        PROTECTED: Override when inheriting.
        """
        pass
    
    @abc.abstractmethod
    def is_finished(self):
        """
        Check if the application has finished.
        
        PROTECTED: Override when inheriting.
        
        Returns
        -------
        finished : bool
            True of the application has finished, false otherwise
        """
        pass
    
    @abc.abstractmethod
    def wait_interval(self):
        """
        The time interval of :func:`is_finished()` calls in the joining
        process.
        
        PROTECTED: Override when inheriting.
        
        Returns
        -------
        interval : float
            Time (in seconds) between calls of :func:`is_finished()` in
            :func:`join()`
        """
        pass
    
    @abc.abstractmethod
    def evaluate(self):
        """
        Evaluate application results. Called in :func:`join()`.
        
        PROTECTED: Override when inheriting.
        """
        pass
    
    def clean_up(self):
        """
        Do clean up work after the application terminates.
        
        PROTECTED: Optionally override when inheriting.
        """
        pass


class AppStateError(Exception):
    """
    Indicate that the application lifecycle was violated.
    """
    pass


class TimeoutError(Exception):
    """
    Indicate that the application's timeout expired.
    """
    pass


class VersionError(Exception):
    """
    Indicate that the application's version is invalid.
    """
    pass


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from distutils.version import Version
import biotite.sequence as seq
import biotite.sequence.phylo as phylo
import biotite.sequence.align as align
from biotite.application import VersionError
from biotite.application.muscle import MuscleApp, Muscle5App
from biotite.application.mafft import MafftApp
from biotite.application.clustalo import ClustalOmegaApp
import numpy as np
import pytest
import shutil
from ..util import is_not_installed


BIN_PATH = {
    MuscleApp : "muscle",
    Muscle5App : "muscle",
    MafftApp : "mafft",
    ClustalOmegaApp: "clustalo"
}


@pytest.fixture
def sequences():
    return [seq.ProteinSequence(string) for string in [
        "BIQTITE",
        "TITANITE",
        "BISMITE",
        "IQLITE"
]]


@pytest.mark.parametrize("app_cls, exp_ali, exp_order",
    [(MuscleApp,
      "BIQT-ITE\n"
      "TITANITE\n"
      "BISM-ITE\n"
      "-IQL-ITE",
      [1, 2, 0, 3]),
     (Muscle5App,
      "BI-QTITE\n"
      "TITANITE\n"
      "BI-SMITE\n"
      "-I-QLITE",
      [0, 3, 1, 2]),
     (MafftApp,
      "-BIQTITE\n"
      "TITANITE\n"
      "-BISMITE\n"
      "--IQLITE",
      [0, 3, 2, 1]),
     (ClustalOmegaApp,
      "-BIQTITE\n"
      "TITANITE\n"
      "-BISMITE\n"
      "--IQLITE",
     [1, 2, 0, 3])]
)
def test_msa(sequences, app_cls, exp_ali, exp_order):
    bin_path = BIN_PATH[app_cls]
    if is_not_installed(bin_path):
        pytest.skip(f"'{bin_path}' is not installed")

    try:
        app = app_cls(sequences)
    except VersionError:
        pytest.skip(f"Invalid software version")
    app.start()
    app.join()
    alignment = app.get_alignment()
    print(alignment)
    order = app.get_alignment_order()
    assert str(alignment) == exp_ali
    assert order.tolist() == exp_order


def test_additional_options(sequences):
    bin_path = BIN_PATH[ClustalOmegaApp]
    if is_not_installed(bin_path):
        pytest.skip(f"'{bin_path}' is not installed")

    app1 = ClustalOmegaApp(sequences)
    app1.start()
    
    app2 = ClustalOmegaApp(sequences)
    app2.add_additional_options(["--full"])
    app2.start()
    
    app1.join()
    app2.join()
    assert "--full" not in app1.get_command()
    assert "--full"     in app2.get_command()
    assert app1.get_alignment() == app2.get_alignment()


@pytest.mark.parametrize("app_cls", [MuscleApp, MafftApp])
def test_custom_substitution_matrix(sequences, app_cls):
    bin_path = BIN_PATH[app_cls]
    if is_not_installed(bin_path):
        pytest.skip(f"'{bin_path}' is not installed")
    
    alph = seq.ProteinSequence.alphabet
    # Strong identity matrix
    score_matrix = np.identity(len(alph)) * 1000
    matrix = align.SubstitutionMatrix(alph, alph, score_matrix)
    exp_ali = (
        "BI-QTITE\n"
        "TITANITE\n"
        "BI-SMITE\n"
        "-I-QLITE"
    )
    try:
        app = app_cls(sequences, matrix=matrix)
    except VersionError:
        pytest.skip(f"Invalid software version")
    app.start()
    app.join()
    alignment = app.get_alignment()
    assert str(alignment) == exp_ali


# Ignore warnings about missing tree output in MUSCLE
@pytest.mark.filterwarnings("ignore")
@pytest.mark.parametrize("app_cls", [MuscleApp, MafftApp])
def test_custom_sequence_type(app_cls):
    bin_path = BIN_PATH[app_cls]
    if is_not_installed(bin_path):
        pytest.skip(f"'{bin_path}' is not installed")
    
    alph = seq.Alphabet(("foo", "bar", 42))
    sequences = [seq.GeneralSequence(alph, sequence) for sequence in [
        ["foo", "bar", 42, "foo",        "foo", 42, 42],
        ["foo",        42, "foo", "bar", "foo", 42, 42],
    ]]
    exp_trace = [
        [ 0,  0],
        [ 1, -1],
        [ 2,  1],
        [ 3,  2],
        [-1,  3],
        [ 4,  4],
        [ 5,  5],
        [ 6,  6],
    ]
    # Strong identity matrix
    score_matrix = np.identity(len(alph))
    score_matrix[score_matrix == 0] = -1000
    score_matrix[score_matrix == 1] = 1000
    matrix = align.SubstitutionMatrix(alph, alph, score_matrix)
    try:
        app = app_cls(sequences, matrix=matrix)
    except VersionError:
        pytest.skip(f"Invalid software version")
    app.start()
    app.join()
    alignment = app.get_alignment()
    assert alignment.sequences == sequences
    assert alignment.trace.tolist() == exp_trace


@pytest.mark.parametrize("app_cls", [MuscleApp, MafftApp, ClustalOmegaApp])
def test_invalid_sequence_type_no_matrix(app_cls):
    """
    A custom substitution matrix is required for normally unsupported
    sequence types.
    """
    bin_path = BIN_PATH[app_cls]
    if is_not_installed(bin_path):
        pytest.skip(f"'{bin_path}' is not installed")
    
    alph = seq.Alphabet(("foo", "bar", 42))
    sequences = [seq.GeneralSequence(alph, sequence) for sequence in [
        ["foo", "bar", 42, "foo",        "foo", 42, 42],
        ["foo",        42, "foo", "bar", "foo", 42, 42],
    ]]
    with pytest.raises(TypeError):
        try:
            app_cls(sequences)
        except VersionError:
            pytest.skip(f"Invalid software version")


@pytest.mark.parametrize("app_cls", [MuscleApp, MafftApp, ClustalOmegaApp])
def test_invalid_sequence_type_unsuitable_alphabet(app_cls):
    """
    The alphabet of the custom sequence type cannot be longer than the
    amino acid alphabet.
    """
    bin_path = BIN_PATH[app_cls]
    if is_not_installed(bin_path):
        pytest.skip(f"'{bin_path}' is not installed")
    
    alph = seq.Alphabet(range(50))
    sequences = [seq.GeneralSequence(alph, sequence) for sequence in [
        [1,2,3],
        [1,2,3],
    ]]
    with pytest.raises(TypeError):
        try:
            app_cls(sequences)
        except VersionError:
            pytest.skip(f"Invalid software version")


def test_invalid_muscle_version(sequences):
    """
    One of `MuscleApp` and `Muscle5App` should raise an error, since one
    is incompatible with the installed version
    """
    bin_path = BIN_PATH[MuscleApp]
    if is_not_installed(bin_path):
        pytest.skip(f"'{bin_path}' is not installed")
    
    if is_not_installed("muscle"):
        pytest.skip(f"'muscle' is not installed")

    with pytest.raises(VersionError):
        MuscleApp(sequences)
        Muscle5App(sequences)


def test_clustalo_matrix(sequences):
    bin_path = BIN_PATH[ClustalOmegaApp]
    if is_not_installed(bin_path):
        pytest.skip(f"'{bin_path}' is not installed")
    
    ref_matrix = [
        [0, 1, 2, 3],
        [1, 0, 1, 2],
        [2, 1, 0, 1],
        [3, 2, 1, 0]
    ]
    app = ClustalOmegaApp(sequences)
    app.full_matrix_calculation()
    app.set_distance_matrix(np.array(ref_matrix))
    app.start()
    app.join()
    test_matrix = app.get_distance_matrix()
    assert np.allclose(ref_matrix, test_matrix)


def test_clustalo_tree(sequences):
    bin_path = BIN_PATH[ClustalOmegaApp]
    if is_not_installed(bin_path):
        pytest.skip(f"'{bin_path}' is not installed")
    
    leaves = [phylo.TreeNode(index=i) for i in range(len(sequences))]
    inter1 = phylo.TreeNode([leaves[0], leaves[1]], [1.0, 1.0])
    inter2 = phylo.TreeNode([leaves[2], leaves[3]], [2.5, 2.5])
    root = phylo.TreeNode([inter1, inter2], [3.5, 2])
    tree = phylo.Tree(root)
    # You cannot simultaneously set and get a tree in ClustalOmega
    # -> Test whether both is possible in separate calls
    app = ClustalOmegaApp(sequences)
    app.set_guide_tree(tree)
    app.start()
    app.join()

    app = ClustalOmegaApp(sequences)
    app.start()
    app.join()
    assert app.get_guide_tree() is not None


def test_mafft_tree(sequences):
    bin_path = BIN_PATH[MafftApp]
    if is_not_installed(bin_path):
        pytest.skip(f"'{bin_path}' is not installed")
    
    app = MafftApp(sequences)
    app.start()
    app.join()
    tree = app.get_guide_tree()
    assert tree is not None


def test_muscle_tree(sequences):
    bin_path = BIN_PATH[MuscleApp]
    if is_not_installed(bin_path):
        pytest.skip(f"'{bin_path}' is not installed")
    
    try:
        app = MuscleApp(sequences)
    except VersionError:
        pytest.skip(f"Invalid software version")
    app.start()
    app.join()
    tree1 = app.get_guide_tree(iteration="kmer")
    tree2 = app.get_guide_tree(iteration="identity")
    assert tree1 is not None
    assert tree2 is not None


def test_muscle5_options(sequences):
    bin_path = BIN_PATH[Muscle5App]
    if is_not_installed(bin_path):
        pytest.skip(f"'{bin_path}' is not installed")
    
    try:
        app = Muscle5App(sequences)
    except VersionError:
        pytest.skip(f"Invalid software version")
    app.use_super5()
    app.set_iterations(2, 100)
    app.set_thread_number(2)
    app.start()

    assert "-super5" in app.get_command()
    assert "-consiters" in app.get_command()
    assert "-refineiters" in app.get_command()
    assert "-threads" in app.get_command()

    app.join()
    assert str(app.get_alignment()) == "BI-QTITE\n" \
                                       "TITANITE\n" \
                                       "BI-SMITE\n" \
                                       "-I-QLITE"


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence"
__author__ = "Patrick Kunzmann"
__all__ = ["find_subsequence", "find_symbol", "find_symbol_first",
           "find_symbol_last"]

import numpy as np


def find_subsequence(sequence, query):
    """
    Find a subsequence in a sequence.
    
    Parameters
    ----------
    sequence : Sequence
        The sequence to find the subsequence in.
    query : Sequence
        The potential subsequence. Its alphabet must extend the
        `sequence` alphabet.
    
    Returns
    -------
    match_indices : ndarray
        The starting indices in `sequence`, where `query` has been
        found. The array is empty if no match has been found.
    
    Raises
    ------
    ValueError
        If the `query` alphabet does not extend the `sequence` alphabet.
    
    Examples
    --------
    
    >>> main_seq = NucleotideSequence("ACTGAATGA")
    >>> sub_seq = NucleotideSequence("TGA")
    >>> print(find_subsequence(main_seq, sub_seq))
    [2 6]
    
    """
    if not sequence.get_alphabet().extends(query.get_alphabet()):
        raise ValueError("The sequences alphabets are not equal")
    match_indices = []
    frame_size = len(query)
    for i in range(len(sequence) - frame_size + 1):
        sub_seq_code = sequence.code[i : i + frame_size]
        if np.array_equal(query.code, sub_seq_code):
            match_indices.append(i)
    return np.array(match_indices)

def find_symbol(sequence, symbol):
    """
    Find a symbol in a sequence.
    
    Parameters
    ----------
    sequence : Sequence
        The sequence to find the symbol in.
    symbol : object
        The symbol to be found in `sequence`.
    
    Returns
    -------
    match_indices : ndarray
        The indices in `sequence`, where `symbol` has been found.
    """
    code = sequence.get_alphabet().encode(symbol)
    return np.where(sequence.code == code)[0]

def find_symbol_first(sequence, symbol):
    """
    Find first occurence of a symbol in a sequence.
    
    Parameters
    ----------
    sequence : Sequence
        The sequence to find the symbol in.
    symbol : object
        The symbol to be found in `sequence`.
    
    Returns
    -------
    first_index : int
        The first index of `symbol` in `sequence`. If `symbol` is not in
        `sequence`, -1 is returned.
    """
    match_i = find_symbol(sequence, symbol)
    if len(match_i) == 0:
        return -1
    return np.min(match_i)
    
def find_symbol_last(sequence, symbol):
    """
    Find last occurence of a symbol in a sequence.
    
    Parameters
    ----------
    sequence : Sequence
        The sequence to find the symbol in.
    symbol : object
        The symbol to be found in `sequence`.
    
    Returns
    -------
    flast_index : int
        The last index of `symbol` in `sequence`. If `symbol` is not in
        `sequence`, -1 is returned.
    """
    match_i = find_symbol(sequence, symbol)
    if len(match_i) == 0:
        return -1
    return np.max(match_i)



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
A subpackage for reading and writing structure related data.

Macromolecular structure files (PDB, PDBx/mmCIF, MMTF, etc.) and
small molecule files (MOL, SDF, etc.) can be used
to load an :class:`AtomArray` or :class:`AtomArrayStack`.

Since the data model for the :class:`AtomArray` and
:class:`AtomArrayStack` class does not support duplicate atoms,
only one *altloc* can be chosen for each atom. Hence, the amount of
atoms may be lower in the atom array (stack) than in respective
structure file.

The recommended format for reading structure files is MMTF.
It has by far the shortest parsing time and file size.
Furthermore, chemical bond information can be read from MMTF files
as :class:`BondList` instances.

Besides the mentioned structure formats, Gromacs trajectory files can be
loaded, if `mdtraj` is installed.
"""

__name__ = "biotite.structure.io"
__author__ = "Patrick Kunzmann"

from .ctab import *
from .general import *
from .trajfile import *


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import numpy as np
import pytest
import biotite.sequence as seq
import biotite.sequence.align as align


@pytest.mark.parametrize("db_entry", [entry for entry
                                      in align.SubstitutionMatrix.list_db()
                                      if entry not in ["NUC","GONNET"]])
def test_matrices(db_entry):
    """
    Test for exceptions when reading matrix files.
    """
    alph1 = seq.ProteinSequence.alphabet
    alph2 = seq.ProteinSequence.alphabet
    matrix = align.SubstitutionMatrix(alph1, alph2, db_entry)

def test_matrix_str():
    """
    Test conversion of substitution matrix to string via a small
    constructed test case.
    """
    alph1 = seq.Alphabet("abc")
    alph2 = seq.Alphabet("def")
    score_matrix = np.arange(9).reshape((3,3))
    matrix = align.SubstitutionMatrix(alph1, alph2, score_matrix)
    assert str(matrix) == "\n".join(
        ["    d   e   f",
         "a   0   1   2",
         "b   3   4   5",
         "c   6   7   8"]
    )


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
A subpackage for fetching data from online databases.

Each subpackage of the database package contains an interface for a
specific database.

All subpackages provide at least two functions:
:func:`search()` is used to search for IDs (e.g. PDB ID) that match the
given search parameters.
The search parameters are usually abstracted by the respective
:class:`Query` objects.
Then the obtained IDs can be given to the :func:`fetch()` function to
download the associated files.
"""

__name__ = "biotite.database"
__author__ = "Patrick Kunzmann"

from .error import *


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from tempfile import TemporaryFile
import glob
import itertools
from os.path import join, splitext
import pytest
from pytest import approx
import numpy as np
import biotite
import biotite.structure.io.gro as gro
import biotite.structure.io.pdb as pdb
from biotite.structure import Atom, array
from ..util import data_dir


def test_get_model_count():
    gro_file = gro.GROFile.read(join(data_dir("structure"), "1l2y.gro"))
    test_model_count = gro_file.get_model_count()
    ref_model_count = gro_file.get_structure().stack_depth()
    assert test_model_count == ref_model_count


@pytest.mark.parametrize(
    "path, model",
    itertools.product(
        glob.glob(join(data_dir("structure"), "*.gro")),
        [None, 1, -1]
    )
)
def test_array_conversion(path, model):
    gro_file = gro.GROFile.read(path)
    array1 = gro_file.get_structure(model=model)
    gro_file = gro.GROFile()
    gro_file.set_structure(array1)
    array2 = gro_file.get_structure(model=model)
    if array1.box is not None:
        assert np.allclose(array1.box, array2.box)
    assert array1.bonds == array2.bonds
    for category in array1.get_annotation_categories():
        assert array1.get_annotation(category).tolist() == \
               array2.get_annotation(category).tolist()
    assert array1.coord.tolist() == array2.coord.tolist()


@pytest.mark.parametrize(
    "path", glob.glob(join(data_dir("structure"), "[!(waterbox)]*.gro"))
)
def test_pdb_consistency(path):
    pdb_path = splitext(path)[0] + ".pdb"
    pdb_file = pdb.PDBFile.read(pdb_path)
    a1 = pdb_file.get_structure(model=1)
    gro_file = gro.GROFile.read(path)
    a2 = gro_file.get_structure(model=1)

    assert a1.array_length() == a2.array_length()

    for category in ["res_id", "res_name", "atom_name"]:
        assert a1.get_annotation(category).tolist() == \
               a2.get_annotation(category).tolist()

    # Mind rounding errors when converting pdb to gro (A -> nm)
    assert a1.coord.flatten().tolist() \
        == approx(a2.coord.flatten().tolist(), abs=1e-2)


@pytest.mark.parametrize(
    "path, model",
    itertools.product(
        glob.glob(join(data_dir("structure"), "*.pdb")),
        [None, 1, -1]
    )
)
def test_pdb_to_gro(path, model):
    """
    Converting stacks between formats should not change data
    """
    # Read in data
    pdb_file = pdb.PDBFile.read(path)
    try:
        a1 = pdb_file.get_structure(model=model)
    except biotite.InvalidFileError:
        if model is None:
            # The file cannot be parsed into an AtomArrayStack,
            # as the models contain different numbers of atoms
            # -> skip this test case
            return
        else:
            raise

    # Save stack as gro
    temp = TemporaryFile("w+")
    gro_file = gro.GROFile()
    gro_file.set_structure(a1)
    gro_file.write(temp)

    # Reload stack from gro
    temp.seek(0)
    gro_file = gro.GROFile.read(temp)
    temp.close()
    a2 = gro_file.get_structure(model=model)

    assert a1.array_length() == a2.array_length()

    for category in ["res_id", "res_name", "atom_name"]:
        assert a1.get_annotation(category).tolist() == \
               a2.get_annotation(category).tolist()

    # Mind rounding errors when converting pdb to gro (A -> nm)
    assert a1.coord.flatten().tolist() \
        == approx(a2.coord.flatten().tolist(), abs=1e-2)


def test_gro_id_overflow():
    # Create an oversized AtomArray where atom_id > 100000 and res_id > 10000
    num_atoms = 100005
    atoms = array([Atom([1,2,3], atom_name="CA", element="C", res_name="X",
                        res_id=i+1) for i in range(num_atoms)])
    atoms.box = np.array([[1,0,0], [0,1,0], [0,0,1]])

    # Write .gro file
    temp = TemporaryFile("w+")
    gro_file = gro.GROFile()
    gro_file.set_structure(atoms)
    gro_file.write(temp)

    # Read .gro file
    temp.seek(0)
    gro_file = gro.GROFile.read(temp)
    temp.close()
    s = gro_file.get_structure()

    assert s.array_length() == num_atoms


def test_gro_no_box():
    """
    .gro file format requires valid box parameters at the end of each
    model. However, if we read such a file in, the resulting object should not
    need to have an assigned box.
    """

    # Create an AtomArray
    atom = Atom([1,2,3], atom_name="CA", element="C", res_name="X", res_id=1)
    atoms = array([atom])

    # Write .gro file
    temp = TemporaryFile("w+")
    gro_file = gro.GROFile()
    gro_file.set_structure(atoms)
    gro_file.write(temp)
    
    # Read in file
    temp.seek(0)
    gro_file = gro.GROFile.read(temp)
    temp.close()
    s = gro_file.get_structure()

    # Assert no box with 0 dimension
    assert s.box is None